c    "Copyright 2010 Prof. Robert Dalrymple, Prof. M. Gomez Gesteira, Dr Benedict Rogers, 
c     Dr Alejandro Crespo, Dr Muthukumar Narayanaswamy, Dr Shan Zou, Dr Andrea Panizzo "
c
c    This file is part of SPHYSICS.
c
c    SPHYSICS is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 3 of the License, or
c    (at your option) any later version.
c
c    SPHYSICS is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program.  If not, see <http://www.gnu.org/licenses/>.

      subroutine celij(j1,j2,kind_p1,ini_kind_p2,lx2,ly2,lz2,id_call)
c
      include 'common.3D'  
      
      if(j2.lt.0)then
        print*
        print*,'ERROR in celij.f'
        print*,'j2.lt.0'
        print*,'j1 ',j1
        print*,'j2 ',j2
        print*,'id_call ',id_call
        print*,'rank_id ',rank_id
        stop
      endif
      
      !do kind_p2=ini_kind_p2,2
      !- Changes looping to include BP-BP interactions
      if(kind_p1.eq.1)then
        kind_p2_start = 1
      else
        kind_p2_start = ini_kind_p2
      endif
      
      do kind_p2=kind_p2_start,2
        if(nc(j2,kind_p2).ne.0) then

        do ii=1,nc(j1,kind_p1)
          i = ibox(j1,kind_p1,ii)
         
          do jj=1,nc(j2,kind_p2)
           j = ibox(j2,kind_p2,jj)
           
           drx = xp(i) - xp(j)
           dry = yp(i) - yp(j)
           drz = zp(i) - zp(j)
            
           call periodicityCorrection
     &          (i,j,drx,dry,drz,lx2,ly2,lz2,id_call)

           rr2 = drx*drx + dry*dry + drz*drz

           if(rr2.lt.rr2_min)then
             if(i_particleType(i).lt.4.and.i_particleType(j).lt.4)then
               goto 641
             end if
              print*
              print*,'ERROR In celij.f: rr2.lt.rr2_min'
              print*,'itime ',itime
              print*,'rank_id ',rank_id
              print*,'j1, j2 ',j1, j2
              print*,'kind_p1, kind_p2 ',kind_p1, kind_p2
              print*,'I, j ',I, j
              print*,'xp(I), yp(I), zp(I) ',xp(i), yp(i), zp(i)
              print*,'xp(j), yp(j), zp(j) ',xp(j), yp(j), zp(j)
              print*,'drx, dry, drz ',drx, dry, drz
              print*,'rr2     ',rr2
              print*,'rr2_min ',rr2_min
              print*,'fourh2 ',fourh2
              print*,'q=sqrt(rr2)/h =',sqrt(rr2)/h
              print*,'i_particleType(i) ',i_particleType(i)
              print*,'i_particleType(j) ',i_particleType(j)
              print*,'pm(I) ',pm(I)
              print*,'pm(j) ',pm(j)
              print*,'up(I), vp(I), wp(I) ',up(i), vp(i), wp(i)
              print*,'up(j), vp(j), wp(j) ',up(j), vp(j), wp(j)
              print*,'rhop(I), rdot(I) ',rhop(i), rdot(i)
              print*,'rhop(j), rdot(j) ',rhop(j), rdot(j)
              call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point
              stop
           end if
           
           if(rr2.lt.fourh2.and.rr2.gt.1.e-18) then
             
             dux = up(i) - up(j)
             duy = vp(i) - vp(j)
             duz = wp(i) - wp(j)

c            Calculating kernel & Normalized Kernel Gradient
             call kernel(drx,dry,drz,i,j,j1,j2,rr2) 
             call kernel_correction(i,j)
             
c
c            ...  average density
c

             robar  = 0.5*( rhop(i) + rhop(j) )
             one_over_rhobar = 2.0/(rhop(i) + rhop(j))
c             
             cbar = 0.5*( cs(i)   + cs(j)   ) 
c
c            ...  inner product rv
c
             dot = drx*dux + dry*duy + drz*duz

c	      Used to calculate the time step due to viscosity

             visc_dt=max(dot/(rr2 + eta2),visc_dt) 
c
c            ...  pressure and viscous force (1992; 3.3)
c
c            pm(j) is mass of particle j
c
c             p_v = pr(i) + pr(j)  !+  pi_visc



c           for tensile correction (Monaghan , JCP.  159 (2000) 290- 311)
c           Only to be activated with cubic spline kernel 
c
            if(index_tensile.eq.1) then

c    ____ Tensile correction
 
              fab=Wab*od_Wdeltap      !NOTE: We'll use a non-normalized
              fab=fab*fab             !kernel to calculate tensile correction
              fab=fab*fab             !It's the (Wab/Wdeltap)**4  of Monaghan's paper

              if (p(i).gt.0) then
                  Ra= 0.01 *pr(i)
              else
                  Ra= 0.2 *abs(pr(i))
              endif

              if (p(j).gt.0) then
                  Rb= 0.01 *pr(j)
              else
                  Rb= 0.2 *abs(pr(j))
              endif

              p_v = p_v+ (Ra+Rb)*fab

            endif
            if(i_particleType(i).eq.4.and.i_particleType(j).eq.4)then !FP-FP

              call approx_RiemannSolver(i,j,drx,dry,drz,rr2,cbar,robar,
     &                           u_a,v_a,w_a,rho_a,p_a,TE_a)
              frx_tmpj = pVol(j)*2.0*frxi
              fry_tmpj = pVol(j)*2.0*fryi
              frz_tmpj = pVol(j)*2.0*frzi
              frx_tmpi = pVol(i)*2.0*frxj
              fry_tmpi = pVol(i)*2.0*fryj
              frz_tmpi = pVol(i)*2.0*frzj
              p_v_x1 = p_a
              p_v_y1 = p_a
              p_v_z1 = p_a
              ax(i) = ax(i) - p_v_x1 * frx_tmpj * one_over_rhop(i)
              ay(i) = ay(i) - p_v_y1 * fry_tmpj * one_over_rhop(i)
              az(i) = az(i) - p_v_z1 * frz_tmpj * one_over_rhop(i)
              ax(j) = ax(j) + p_v_x1 * frx_tmpi * one_over_rhop(j)
              ay(j) = ay(j) + p_v_y1 * fry_tmpi * one_over_rhop(j)
              az(j) = az(j) + p_v_z1 * frz_tmpi * one_over_rhop(j)

              call gradients_calc(i,j,dux,duy,duz)               

              call viscosity(dot,drx,dry,drz,dux,duy,duz,rr2,
     +               cbar,robar,one_over_rhobar,i,j,j1,j2,term2i,term2j)

c  ...  density acceleration 
c       
              
              dot2ai = (u_a - up(j))*frx_tmpi
     &               + (v_a - vp(j))*fry_tmpi + (w_a - wp(j))*frz_tmpi
              dot2aj = (u_a - up(i))*frx_tmpj
     &               + (v_a - vp(i))*fry_tmpj + (w_a - wp(i))*frz_tmpj
              ar(i) = ar(i) - rhop(i)*dot2aj
              ar(j) = ar(j) + rhop(j)*dot2ai

c  ...  Thermal Energy
c
              aTE(i) = aTE(i) - p_a*dot2aj * one_over_rhop(i)
              aTE(j) = aTE(j) + p_a*dot2ai * one_over_rhop(j)
c
c  ...  XSPH correction (Monaghan 1994;  J. Comp. Phys. 110. Formula 2.6)
c
              pmj_Wab_over_rhobar = pm(j)*Wab*one_over_rhobar
              ux(i) = ux(i) - dux*pmj_Wab_over_rhobar  !pm(j) * dux * Wab / robar
              vx(i) = vx(i) - duy*pmj_Wab_over_rhobar  !pm(j) * duy * Wab / robar
              wx(i) = wx(i) - duz*pmj_Wab_over_rhobar  !pm(j) * duz * Wab / robar

              pmi_Wab_over_rhobar = pm(i)*Wab*one_over_rhobar
              ux(j) = ux(j) + dux*pmi_Wab_over_rhobar   !pm(i) * dux * Wab / robar
              vx(j) = vx(j) + duy*pmi_Wab_over_rhobar   !pm(i) * duy * Wab / robar
              wx(j) = wx(j) + duz*pmi_Wab_over_rhobar   !pm(i) * duz * Wab / robar

c  ...  Vorticity calculation

              if(ipoute.eq.1.and.i_vort.eq.1.and.
     &          i_particleType(i).gt.3.and.i_particleType(j).gt.3)then  !FP-FP
                  call vorticity_calc(i,j,dux,duy,duz)
              endif  
		   			
            else if(i_particleType(i).eq.4.and.
     &              i_particleType(j).lt.4)then  !FP-BP,   j is boundary particle
              call MonaghanBC(i,j,drx,dry,drz,dot,dux,duy,duz,
     &                        fxbMON,fybMON,fzbMON)   
              ax(i) = ax(i) + fxbMON
              ay(i) = ay(i) + fybMON
              az(i) = az(i) + fzbMON
              temp = pVol(i)*visc_wall*
     &           ((drx*frxi+dry*fryi+drz*frzi)/(rr2 + eta2))
              ax(i) = ax(i) + temp*dux
              ay(i) = ay(i) + temp*duy
              az(i) = az(i) + temp*duz

              if(i_particleType(j).eq.3)then  !Opposite Forces on free-moving object (Not per unit mass, see subroutine whole_body_motion)
                ax(j) = ax(j) - temp*dux*pm(i)
                ay(j) = ay(j) - temp*duy*pm(i)
                az(j) = az(j) - temp*duz*pm(i)
                ax(j) = ax(j) - fxbMON*pm(i)
                ay(j) = ay(j) - fybMON*pm(i)
                az(j) = az(j) - fzbMON*pm(i)
              end if

            else if(i_particleType(j).eq.4.and.
     &              i_particleType(i).lt.4)then  !FP-BP
              call MonaghanBC(j,i,-drx,-dry,-drz,dot,-dux,-duy,-duz,
     &                        fxbMON,fybMON,fzbMON)   
              ax(j) = ax(j) + fxbMON
              ay(j) = ay(j) + fybMON
              az(j) = az(j) + fzbMON
              temp = pVol(j)*visc_wall*
     &           ((drx*frxj+dry*fryj+drz*frzj)/(rr2 + eta2))
              ax(j) = ax(j) - temp*dux
              ay(j) = ay(j) - temp*duy
              az(j) = az(j) - temp*duz

              if(i_particleType(i).eq.3)then  !Opposite Forces on free-moving object (Not per unit mass, see subroutine whole_body_motion)
                ax(i) = ax(i) + temp*dux*pm(j)
                ay(i) = ay(i) + temp*duy*pm(j)
                az(i) = az(i) + temp*duz*pm(j)
                ax(i) = ax(i) - fxbMON*pm(j)
                ay(i) = ay(i) - fybMON*pm(j)
                az(i) = az(i) - fzbMON*pm(j)
              end if

            else if(i_particleType(i).lt.4.and.
     &              i_particleType(j).lt.4)then !BP-BP
              if(i_particleType(i).gt.2)then !i is free-moving BP
                if(i_particleType(j).gt.2)then  !j is free-moving BP
                  if(num_FB.gt.1)then
                    !Loop to prevent BP within a Floating Body interacting with a BP within the same Floating Body
                    i_num_FB = 1
                    i_BP_BP_interaction = 0
                    i_absolute = iabsolute_BP_Pointer_Index(i)
                    j_absolute = iabsolute_BP_Pointer_Index(j)
                    id_num_FB_i = i_FB_Pointer_Info(i_absolute)
                    id_num_FB_j = i_FB_Pointer_Info(j_absolute)
                    do while (i_BP_BP_interaction.lt.1)
                      if(i_num_FB.lt.num_FB)then
                        i_num_FB = i_num_FB + 1
                        if(id_num_FB_i.ne.id_num_FB_j)then
                          call MonaghanBC
     &                         (i,j,drx,dry,drz,dot,dux,duy,duz,
     &                          fxbMON,fybMON,fzbMON)
                          ax(i) = ax(i) + fxbMON*pm(j)
                          ay(i) = ay(i) + fybMON*pm(j)
                          az(i) = az(i) + fzbMON*pm(j)
                          ax(j) = ax(j) - fxbMON*pm(j)
                          ay(j) = ay(j) - fybMON*pm(j)
                          az(j) = az(j) - fzbMON*pm(j)
                          i_BP_BP_interaction = 1
                        endif
                      else
                        i_BP_BP_interaction = 1
                      endif
                    enddo
                  endif
                else  !j is prescribed BP
                  call MonaghanBC(i,j,drx,dry,drz,dot,dux,duy,duz,
     &                            fxbMON,fybMON,fzbMON)
                  ax(i) = ax(i) + fxbMON*pm(j)
                  ay(i) = ay(i) + fybMON*pm(j)
                  az(i) = az(i) + fzbMON*pm(j)
                end if    !End of:  if(j.gt.nbfm)then
              else if(i_particleType(j).eq.3)then   !j is free-moving BP, i is prescibed BP
                call MonaghanBC(j,i,-drx,-dry,-drz,dot,-dux,-duy,-duz,
     &                          fxbMON,fybMON,fzbMON)
                ax(j) = ax(j) + fxbMON*pm(i)
                ay(j) = ay(j) + fybMON*pm(i)
                az(j) = az(j) + fzbMON*pm(i)
              endif   !End of:  if(i.gt.nbfm)then
             endif  ! Interaction fluid-fluid or fluid-boundary
		  
           endif ! if q<2
          enddo ! Box jj
         enddo ! Box ii
641     endif  ! Box jj is not empty
      enddo   ! Kind of particle

      end