c    "Copyright 2010 Prof. Robert Dalrymple, Prof. M. Gomez Gesteira, Dr Benedict Rogers, 
c     Dr Alejandro Crespo, Dr Muthukumar Narayanaswamy, Dr Shan Zou, Dr Andrea Panizzo "
c
c    This file is part of SPHYSICS.
c
c    SPHYSICS is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 3 of the License, or
c    (at your option) any later version.
c
c    SPHYSICS is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program.  If not, see <http://www.gnu.org/licenses/>.

c
c*********************************************************************
c   !BDR 04/09/06
c   This subroutine performs the following operations:
c
c     (i)   Checks to see if particles are leaving the subdomain of 
c           the processor (i.e. the rank)
c     (ii)  Transfers the particles and all their properties to the
c           adjacent process
c     (iii) On each processor, the ip index system is then reorganised
c           for the beginning of the next timestep
c
c     Note, this subroutine re-organises the index pointer system but does not
c     reorganise the data array

      subroutine index_check_NorthSouth

      include 'common.3D'
      
      real(kind=kind_XYZ) xp_temp,yp_temp,zp_temp
      
      double precision dxp_dble, dyp_dble, dzp_dble

      integer iBPs_leaving_North(nm_max)  , iBPs_leaving_South(nm_max)
      integer iBPs_entering_North(nm_max) , iBPs_entering_South(nm_max)
      integer iFPs_leaving_North(nm_max)  , iFPs_leaving_South(nm_max)
      integer iFPs_entering_North(nm_max) , iFPs_entering_South(nm_max)

      !Declarations of Particles leaving/entering a processor

      !-- Derived DataTypes --
      real(kind=kind_XYZ) P_xyz_data_leaving_North(nm_max,3)
      real(kind=kind_XYZ) P_xyz_data_leaving_South(nm_max,3)
      real(kind=kind_XYZ) P_xyz_data_entering_North(nm_max,3)
      real(kind=kind_XYZ) P_xyz_data_entering_South(nm_max,3)
      real P_data_leaving_North(nm_max,31)
      real P_data_leaving_South(nm_max,31)
      real P_data_entering_North(nm_max,31)
      real P_data_entering_South(nm_max,31)

      real rhop_sum_data_leaving_North(nm_max)
      real rhop_sum_data_leaving_South(nm_max)
      real rhop_sum_data_entering_North(nm_max)
      real rhop_sum_data_entering_South(nm_max)
      
      real P_verlet_data_leaving_North(nm_max,5)
      real P_verlet_data_leaving_South(nm_max,5)
      real P_verlet_data_entering_North(nm_max,5)
      real P_verlet_data_entering_South(nm_max,5)
      
      integer i_ABP_leaving_North(nm_max)
      integer i_ABP_leaving_South(nm_max)
      integer i_ABP_entering_North(nm_max)
      integer i_ABP_entering_South(nm_max)

      integer np_nums_leaving_North(5), np_nums_entering_South(5)
      integer np_nums_leaving_South(5), np_nums_entering_North(5)
      
      itime_check = 10e8     !0     !0    !10e8     !
      i_rank_check = 0

!       if(rank_id.ge.i_rank_check.and.itime.ge.itime_check)then
!         print*
!         print*,'BEGINNING OF index_check_NorthSouth.f'
!         print*,'itime, rank_id ',itime, rank_id
!         print*
!       end if
      
      !Artificially move the particles from one proc to another
      !if(rank_id.eq.0)then
      ! do ipcount=1,np_local
      !  if(ipcount.le.nb_local)then
      !    i = ip_index_BP_local(ipcount)
      !  else
      !    i = ip_index_FP_local(ipcount-nb_local)
      !  end if
      !    if(xp(i).gt.2.66)then
      !      xp(i) = xp(i) + 0.05
      !    endif
      ! enddo
      !endif


c      call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point

     
c      !-- Debug Code------
c      if(itime.ge.itime_check.and.rank_id.eq.i_rank_check)then
c        print*
c        print*,'Index_check'
c        print*,'itime ',itime
c        print*
c        print*,'rank_id ',rank_id
c        print*,'nbf_local        ',nbf_local
c        print*,'nbfm_local  ',nbfm_local
c        print*,'nbfreem_local      ',nbfreem_local
c        print*,'nb_local         ',nb_local
c        print*,'np_local         ',np_local
c        print*
c        !print*,'Cartesian Topology'        
c        nbf_index       = 0
c        nbfm_index = 0
c        nbfreem_index     = 0
c      end if
c      if(itime.ge.itime_check)then
c        do i=0,n_procs-1
c          if(i.eq.-1-rank_id)then
c            write(*,*),'rank_id, north, south ',rank_id, north, south 
c          end if
c          call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point
c        end do
c        
c        call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point
c      end if
c      !-------------------

      nb_leaving_North = 0
      nb_leaving_South = 0
      np_leaving_North = 0
      np_leaving_South = 0
      nb_entering_North = 0
      nb_entering_South = 0
      np_entering_North = 0
      np_entering_South = 0

      no_of_bfPs_leaving_North = 0
      no_of_bfPs_leaving_South = 0
      no_of_forcedmBPs_leaving_North = 0      !MBP=Forced Moving Boundary Particles
      no_of_forcedmBPs_leaving_South = 0      !MBP=Forced Moving Boundary Particles
      no_of_freemBPs_leaving_North = 0      !MBP=Free Moving Boundary Particles
      no_of_freemBPs_leaving_South = 0      !MBP=Free Moving Boundary Particles
      no_of_bfPs_entering_North = 0
      no_of_bfPs_entering_South = 0
      no_of_forcedmBPs_entering_North = 0     !MBP=Forced Moving Boundary Particles
      no_of_forcedmBPs_entering_South = 0     !MBP=Forced Moving Boundary Particles
      no_of_freemBPs_entering_North = 0     !MBP=Free Moving Boundary Particles
      no_of_freemBPs_entering_South = 0     !MBP=Free Moving Boundary Particles


      no_of_BPs_leaving_North = 0
      no_of_BPs_leaving_South = 0
      no_of_BPs_entering_North = 0
      no_of_BPs_entering_South = 0

      no_of_FPs_leaving_North = 0
      no_of_FPs_leaving_South = 0
      no_of_FPs_entering_North = 0
      no_of_FPs_entering_South = 0

      do i=1,nm_max
        iBPs_leaving_North(i) = 0
        iBPs_leaving_South(i) = 0
        iFPs_leaving_North(i) = 0
        iFPs_leaving_South(i) = 0

        iBPs_entering_North(i) = 0
        iBPs_entering_South(i) = 0
        iFPs_entering_North(i) = 0
        iFPs_entering_South(i) = 0
      end do
      
      if(i_densityFilter.gt.0.and.
     &        itime/ndt_FilterPerform*ndt_FilterPerform.eq.itime)then   !ndt_FilterPerform set in getdata.f 
        i_filterTransfer_Perform = 1
      else
        i_filterTransfer_Perform = 0
      endif
      
      if(i_MPI_Adapt.eq.1.and.n_procs_y.gt.1)then
        if(itime/n_itime_Adapt*n_itime_Adapt.eq.itime
     &                      .or.itime.eq.itime_init)then
          call Adapt_Check_NorthSouth
        endif
      endif
c
c      itop_free_index_temp = itop_free_index  ! BDR itop_free_index
c                                             ! points to the last member in list of free
c                                             ! array locations (ifree_index)

      !- Set ymax_rank_check to account for periodicity -ymax_container_double
      if((iCartTop_periodicity(2).or.i_periodicOBs(2).eq.1).and.
     &   rank_idy.eq.n_procs_y-1)then
        ymax_rank_check = ymax_container  !_double
      else
        ymax_rank_check = ymax_rank
      endif

c     ***  check to see if particles leave the domain  ***
      !- Boundary Particles -
      !- Fixed Boundary Particles -
      do ipcount = 1,nbf_local
        i=ip_index_BP_local_1(ipcount)
        if(yp(i).gt.ymax_rank_check)then  !Check North boundary
          no_of_BPs_leaving_North = no_of_BPs_leaving_North + 1
          iBPs_leaving_North(no_of_BPs_leaving_North) = i
          iBPs_top_free_index = iBPs_top_free_index +1
          iBPs_free_index(iBPs_top_free_index) = i
          iarray_loc_in_use(i) = 0
          no_of_bfPs_leaving_North = no_of_bfPs_leaving_North +1
        else if(yp(i).lt.ymin_rank)then  !Check South boundary
          no_of_BPs_leaving_South = no_of_BPs_leaving_South + 1
          iBPs_leaving_South(no_of_BPs_leaving_South) = i
          iBPs_top_free_index = iBPs_top_free_index +1
          iBPs_free_index(iBPs_top_free_index) = i
          iarray_loc_in_use(i) = 0
          no_of_bfPs_leaving_South = no_of_bfPs_leaving_South +1
        end if
      end do
      !- Forced Boundary Particles -
      do ipcount = nbf_local + 1,nbfm_local
        i=ip_index_BP_local_2(ipcount - nbf_local)
        if(yp(i).gt.ymax_rank_check)then  !Check North boundary
          no_of_BPs_leaving_North = no_of_BPs_leaving_North + 1
          iBPs_leaving_North(no_of_BPs_leaving_North) = i
          iBPs_top_free_index = iBPs_top_free_index +1
          iBPs_free_index(iBPs_top_free_index) = i
          iarray_loc_in_use(i) = 0
          no_of_forcedmBPs_leaving_North = 
     &     no_of_forcedmBPs_leaving_North+1
        else if(yp(i).lt.ymin_rank)then  !Check South boundary
          no_of_BPs_leaving_South = no_of_BPs_leaving_South + 1
          iBPs_leaving_South(no_of_BPs_leaving_South) = i
          iBPs_top_free_index = iBPs_top_free_index +1
          iBPs_free_index(iBPs_top_free_index) = i
          iarray_loc_in_use(i) = 0
          no_of_forcedmBPs_leaving_South = 
     &     no_of_forcedmBPs_leaving_South+1
        end if
      end do
      !- Free-Moving Boundary Particles -
      do ipcount = nbfm_local + 1,nb_local
        i=ip_index_BP_local_3(ipcount - nbfm_local)
        if(yp(i).gt.ymax_rank_check)then  !Check North boundary
          no_of_BPs_leaving_North = no_of_BPs_leaving_North + 1
          iBPs_leaving_North(no_of_BPs_leaving_North) = i
          iBPs_top_free_index = iBPs_top_free_index +1
          iBPs_free_index(iBPs_top_free_index) = i
          iarray_loc_in_use(i) = 0
          no_of_freemBPs_leaving_North = no_of_freemBPs_leaving_North +1
        else if(yp(i).lt.ymin_rank)then  !Check South boundary
          no_of_BPs_leaving_South = no_of_BPs_leaving_South + 1
          iBPs_leaving_South(no_of_BPs_leaving_South) = i
          iBPs_top_free_index = iBPs_top_free_index +1
          iBPs_free_index(iBPs_top_free_index) = i
          iarray_loc_in_use(i) = 0
          no_of_freemBPs_leaving_South = no_of_freemBPs_leaving_South +1
        end if
      end do          
      !- Fluid Particles -
      do ipcount = nb_local + 1,np_local
        i=ip_index_FP_local(ipcount - nb_local)
        if(yp(i).gt.ymax_rank_check)then  !Check North boundary
          no_of_FPs_leaving_North = no_of_FPs_leaving_North + 1
          iFPs_leaving_North(no_of_FPs_leaving_North) = i
          iFPs_top_free_index = iFPs_top_free_index +1
          iFPs_free_index(iFPs_top_free_index) = i
          iarray_loc_in_use(i) = 0
        else if(yp(i).lt.ymin_rank)then  !Check South boundary
          no_of_FPs_leaving_South = no_of_FPs_leaving_South + 1
          iFPs_leaving_South(no_of_FPs_leaving_South) = i
          iFPs_top_free_index = iFPs_top_free_index +1
          iFPs_free_index(iFPs_top_free_index) = i
          iarray_loc_in_use(i) = 0
        end if

      end do

      nb_leaving_North = no_of_BPs_leaving_North
      np_leaving_North = no_of_BPs_leaving_North+no_of_FPs_leaving_North
      nb_leaving_South = no_of_BPs_leaving_South
      np_leaving_South = no_of_BPs_leaving_South+no_of_FPs_leaving_South

c      !-- Debug Code------
c      if(itime.ge.itime_check)then
c        do i=0,n_procs-1
c          if(rank_id.eq.i)then
c            print*
c            print*,'itime ',itime
c            write(*,*),'rank_id ', rank_id
c            write(*,*),'nb_leaving_North, nb_leaving_South ',
c     &                  nb_leaving_North, nb_leaving_South 
c            write(*,*),'np_leaving_North, np_leaving_South ',
c     &                  np_leaving_North, np_leaving_South 
c            print*,'no_of_bfPs_leaving_North ',no_of_bfPs_leaving_North
c            print*,'no_of_bfPs_leaving_South ',no_of_bfPs_leaving_South
c            print*,'no_of_forcedmBPs_leaving_North ',
c     &              no_of_forcedmBPs_leaving_North
c            print*,'no_of_forcedmBPs_leaving_South ',
c     &              no_of_forcedmBPs_leaving_South
c            print*,'no_of_freemBPs_leaving_North ',no_of_freemBPs_leaving_North
c            print*,'no_of_freemBPs_leaving_South ',no_of_freemBPs_leaving_South
c            print*,'no_of_FPs_leaving_North ',no_of_FPs_leaving_North
c            print*,'no_of_FPs_leaving_South ',no_of_FPs_leaving_South
c            print*,'nbf_local        ',nbf_local
c            print*,'nbfm_local  ',nbfm_local
c            print*,'nbfreem_local      ',nbfreem_local
c            print*,'nb_local         ',nb_local
c            print*,'np_local         ',np_local
c          end if
c          call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point
c        end do
c      end if
c      !----------------------

      if(nb_leaving_North.gt.nm_max.or.nb_leaving_South.gt.nm_max.or.
     &   np_leaving_North.gt.nm_max.or.np_leaving_South.gt.nm_max.or.
     &   nb_leaving_North.lt.0.or.nb_leaving_South.lt.0.or.
     &   np_leaving_North.lt.0.or.np_leaving_South.lt.0)then
        print*
        print*,'Index_check.f'      
        print*,'n#_leaving_##st.gt.nm_max OR'
        print*,'n#_leaving_##st.lt.0'
        print*,'itime ',itime
        write(*,*),'rank_id ', rank_id
        print*,'mn_max ',nm_max
        write(*,*),'nb_leaving_North, nb_leaving_South ',
     &              nb_leaving_North, nb_leaving_South 
        write(*,*),'np_leaving_North, np_leaving_South ',
     &              np_leaving_North, np_leaving_South 
        print*,'no_of_bfPs_leaving_North ',no_of_bfPs_leaving_North
        print*,'no_of_bfPs_leaving_South ',no_of_bfPs_leaving_South
        print*,'no_of_forcedmBPs_leaving_North ',
     &          no_of_forcedmBPs_leaving_North
        print*,'no_of_forcedmBPs_leaving_South ',
     &          no_of_forcedmBPs_leaving_South
        print*,'no_of_freemBPs_leaving_North',
     &          no_of_freemBPs_leaving_North
        print*,'no_of_freemBPs_leaving_South',
     &          no_of_freemBPs_leaving_South
        print*,'no_of_FPs_leaving_North ',no_of_FPs_leaving_North
        print*,'no_of_FPs_leaving_South ',no_of_FPs_leaving_South
        print*,'nbf_local        ',nbf_local
        print*,'nbfm_local  ',nbfm_local
        print*,'nbfreem_local      ',nbfreem_local
        print*,'nb_local         ',nb_local
        print*,'np_local         ',np_local
        stop
      end if

      !call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point

      !--- PACK DATA TO BE SENT EAST ---
c      !-- Debug Code------
c      if(rank_id.eq.i_rank_check.and.itime.ge.itime_check)then
c        print*
c        print*,'Packing Particle Data leaving North'
c        print*,'rank_id ',rank_id
c        print*,'nb_leaving_North ',nb_leaving_North
c        print*,'np_leaving_North ',np_leaving_North
c        print*,'no_of_bfPs_leaving_North      ',no_of_bfPs_leaving_North
c        print*,'no_of_forcedmBPs_leaving_North ',
c     &          no_of_forcedmBPs_leaving_North
c        print*,'no_of_freemBPs_leaving_North      ',no_of_freemBPs_leaving_North
c        print*,'no_of_BPs_leaving_North ',no_of_BPs_leaving_North
c        print*,'no_of_FPs_leaving_North ',no_of_FPs_leaving_North
c        print*,' ii,   i, xp_leaving_North(ii), zp_leaving_North(ii) '
c      end if
      do ii=1,np_leaving_North
        if(ii.le.nb_leaving_North)then        
          i = iBPs_leaving_North(ii)
        else
          i = iFPs_leaving_North(ii-nb_leaving_North)
        end if        
        P_xyz_data_leaving_North(ii,1)  = xp(i)
        P_xyz_data_leaving_North(ii,2)  = yp(i)        
        P_xyz_data_leaving_North(ii,3)  = zp(i)        
        P_data_leaving_North(ii,1)  = up(i)
        P_data_leaving_North(ii,2)  = vp(i)        
        P_data_leaving_North(ii,3)  = wp(i)        
        P_data_leaving_North(ii,4)  = rhop(i)
        P_data_leaving_North(ii,5)  = p(i)        
        P_data_leaving_North(ii,6)  = pm(i)        
        P_data_leaving_North(ii,7)  = pVol(i)        
        P_data_leaving_North(ii,8)  = cs(i)        
        P_data_leaving_North(ii,9)  = sum_wab(i)
        P_data_leaving_North(ii,10) = tau_xx_uno_rho2(i)
        P_data_leaving_North(ii,11) = tau_xy_uno_rho2(i)
        P_data_leaving_North(ii,12) = tau_xz_uno_rho2(i)
        P_data_leaving_North(ii,13) = tau_yy_uno_rho2(i)
        P_data_leaving_North(ii,14) = tau_yz_uno_rho2(i)
        P_data_leaving_North(ii,15) = tau_zz_uno_rho2(i)
        P_data_leaving_North(ii,16) = TEp(i)
        P_data_leaving_North(ii,17) = dudx_CSPH(i)
        P_data_leaving_North(ii,18) = dudy_CSPH(i)
        P_data_leaving_North(ii,19) = dudz_CSPH(i)
        P_data_leaving_North(ii,20) = dvdx_CSPH(i)
        P_data_leaving_North(ii,21) = dvdy_CSPH(i)
        P_data_leaving_North(ii,22) = dvdz_CSPH(i)
        P_data_leaving_North(ii,23) = dwdx_CSPH(i)
        P_data_leaving_North(ii,24) = dwdy_CSPH(i)
        P_data_leaving_North(ii,25) = dwdz_CSPH(i)
        P_data_leaving_North(ii,26) = drhodx_CSPH(i)
        P_data_leaving_North(ii,27) = drhody_CSPH(i)
        P_data_leaving_North(ii,28) = drhodz_CSPH(i)
        P_data_leaving_North(ii,29) = dTEdx_CSPH(i)
        P_data_leaving_North(ii,30) = dTEdy_CSPH(i)
        P_data_leaving_North(ii,31) = dTEdz_CSPH(i)
        if(i_filterTransfer_Perform.eq.1)then
          rhop_sum_data_leaving_North(ii) = rhop_sum(i)
        endif
        !Updating Normals Data
        if(ii.le.nb_leaving_North)then
          i_ABP_leaving_North(ii) = iabsolute_BP_Pointer_Index(i)    
        end if
        !- To be used with Verlet algorithm -
        if(i_algorithm.eq.2) then 
          P_verlet_data_leaving_North(ii,1) = uo(i)
          P_verlet_data_leaving_North(ii,2) = vo(i)
          P_verlet_data_leaving_North(ii,3) = wo(i)
          P_verlet_data_leaving_North(ii,4) = rhoo(i)
          P_verlet_data_leaving_North(ii,5) = TEo(i)
        endif

c        !-- Debug Code------
c        if(rank_id.eq.2.and.itime.ge.itime_check)then
c          print*,'ii, i',ii, i
c          print*,'xp,zp_leaving_North(ii) ',
c     &    P_xyz_Data_leaving_North(ii,1),
c     &    P_xyz_Data_leaving_North(ii,2),
c     &    P_xyz_Data_leaving_North(ii,3)
c          print*,'i_particleType(i) ',i_particleType(i)
c          if(ii.le.nb_leaving_North)then
c            print*,'iabsolute_BP_Pointer_Index(i) ',
c     &              i_ABP_leaving_North(ii)
c          end if
c        end if
c        !-------------------

      end do

      !--- PACK DATA TO BE SENT WEST ---
c      !-- Debug Code------
c      if(rank_id.eq.3.and.itime.ge.itime_check)then
c        print*
c        print*,'Packing Particle Data leaving South'
c        print*,'rank_id ',rank_id
c        print*,'nb_leaving_South ',nb_leaving_South
c        print*,'np_leaving_South ',np_leaving_South
c        print*,'no_of_bfPs_leaving_South      ',no_of_bfPs_leaving_South
c        print*,'no_of_forcedmBPs_leaving_South ',
c     &          no_of_forcedmBPs_leaving_South
c        print*,'no_of_freemBPs_leaving_South      ',no_of_freemBPs_leaving_South
c        print*,'no_of_BPs_leaving_South ',no_of_BPs_leaving_South
c        print*,'no_of_FPs_leaving_South ',no_of_FPs_leaving_South
c        print*,' ii,   i, xp_leaving_South(ii), zp_leaving_South(ii) '
c      end if
c      !--------------------
      do ii=1,np_leaving_South
        if(ii.le.nb_leaving_South)then        
          i = iBPs_leaving_South(ii)
        else
          i = iFPs_leaving_South(ii-nb_leaving_South)
        end if        
        P_xyz_data_leaving_South(ii,1)  = xp(i)
        P_xyz_data_leaving_South(ii,2)  = yp(i)        
        P_xyz_data_leaving_South(ii,3)  = zp(i)        
        P_data_leaving_South(ii,1)  = up(i)
        P_data_leaving_South(ii,2)  = vp(i)
        P_data_leaving_South(ii,3)  = wp(i)        
        P_data_leaving_South(ii,4)  = rhop(i)
        P_data_leaving_South(ii,5)  = p(i)        
        P_data_leaving_South(ii,6)  = pm(i)        
        P_data_leaving_South(ii,7)  = pVol(i)        
        P_data_leaving_South(ii,8)  = cs(i)        
        P_data_leaving_South(ii,9)  = sum_wab(i)
        P_data_leaving_South(ii,10) = tau_xx_uno_rho2(i)
        P_data_leaving_South(ii,11) = tau_xy_uno_rho2(i)
        P_data_leaving_South(ii,12) = tau_xz_uno_rho2(i)
        P_data_leaving_South(ii,13) = tau_yy_uno_rho2(i)
        P_data_leaving_South(ii,14) = tau_yz_uno_rho2(i)
        P_data_leaving_South(ii,15) = tau_zz_uno_rho2(i)
        P_data_leaving_South(ii,16) = TEp(i)
        P_data_leaving_South(ii,17) = dudx_CSPH(i)
        P_data_leaving_South(ii,18) = dudy_CSPH(i)
        P_data_leaving_South(ii,19) = dudz_CSPH(i)
        P_data_leaving_South(ii,20) = dvdx_CSPH(i)
        P_data_leaving_South(ii,21) = dvdy_CSPH(i)
        P_data_leaving_South(ii,22) = dvdz_CSPH(i)
        P_data_leaving_South(ii,23) = dwdx_CSPH(i)
        P_data_leaving_South(ii,24) = dwdy_CSPH(i)
        P_data_leaving_South(ii,25) = dwdz_CSPH(i)
        P_data_leaving_South(ii,26) = drhodx_CSPH(i)
        P_data_leaving_South(ii,27) = drhody_CSPH(i)
        P_data_leaving_South(ii,28) = drhodz_CSPH(i)
        P_data_leaving_South(ii,29) = dTEdx_CSPH(i)
        P_data_leaving_South(ii,30) = dTEdy_CSPH(i)
        P_data_leaving_South(ii,31) = dTEdz_CSPH(i)
        if(i_filterTransfer_Perform.eq.1)then
          rhop_sum_data_leaving_South(ii) = rhop_sum(i)
        endif
        !Updating Normals Data
        if(ii.le.nb_leaving_South)then
          i_ABP_leaving_South(ii) = iabsolute_BP_Pointer_Index(i)    
        end if
        !- To be used with Verlet algorithm -
        if(i_algorithm.eq.2) then 
          P_verlet_data_leaving_South(ii,1) = uo(i)
          P_verlet_data_leaving_South(ii,2) = vo(i)
          P_verlet_data_leaving_South(ii,3) = wo(i)
          P_verlet_data_leaving_South(ii,4) = rhoo(i)
          P_verlet_data_leaving_South(ii,5) = TEo(i)
        endif

c        !-- Debug Code------
c        if(rank_id.eq.2.and.itime.ge.itime_check)then   !if(rank_id.eq.-10)then
c          print*,'ii, i',ii, i
c          print*,'xp_leaving_South(ii), zp_leaving_South(ii) ',
c     &    P_xyz_Data_leaving_South(ii,1), 
c     &    P_xyz_Data_leaving_South(ii,2), 
c     &    P_xyz_Data_leaving_South(ii,3)
c          print*,'i_particleType(i) ',i_particleType(i)
c          if(ii.le.nb_leaving_South)then
c            print*,'iabsolute_BP_Pointer_Index(i) ',
c     &              i_ABP_leaving_South(ii)
c          end if
c        end if
c
c        if(rank_id.eq.2.and.itime.ge.itime_check)then   !if(rank_id.eq.-10)then
c          if(i.eq.98)then
c            print*,'particle i= ',i,' leaving rank_id ',rank_id
c            print*,'ii leaving north ',ii
c            print*,'P_xyz_data_leaving_South(ii,1)  = xp(i)          ',
c     &              P_xyz_data_leaving_South(ii,1)
c            print*,'P_xyz_data_leaving_South(ii, )  = yp(i)          ',
c     &              P_xyz_data_leaving_South(ii, )
c            print*,'P_xyz_data_leaving_South(ii,2)  = zp(i)          ',
c     &              P_xyz_data_leaving_South(ii,2)
c            print*,'P_data_leaving_South(ii,1)  = up(i)              ',
c     &              P_data_leaving_South(ii,1)
c            print*,'P_data_leaving_South(ii, )  = vp(i)              ',
c     &              P_data_leaving_South(ii, )
c            print*,'P_data_leaving_South(ii,2)  = wp(i)              ',
c     &              P_data_leaving_South(ii,2)
c            print*,'P_data_leaving_South(ii,3)  = rhop(i)            ',
c     &              P_data_leaving_South(ii,3)
c            print*,'P_data_leaving_South(ii,4)  = p(i)               ',
c     &              P_data_leaving_South(ii,4)
c            print*,'P_data_leaving_South(ii,5)  = pm(i)              ',
c     &              P_data_leaving_South(ii,5)
c            print*,'P_data_leaving_South(ii,6)  = pVol(i)            ',
c     &              P_data_leaving_South(ii,6)
c            print*,'P_data_leaving_South(ii,7)  = cs(i)              ',
c     &              P_data_leaving_South(ii,7)
c            print*,'P_data_leaving_South(ii,8) = sum_wab(i)          ',
c     &              P_data_leaving_South(ii,8)
c            print*,'P_data_leaving_South(ii,9)  = tau_xx_uno_rho2(i) ',
c     &              P_data_leaving_South(ii,9)
c            print*,'P_data_leaving_South(ii, )  = tau_xy_uno_rho2(i) ',
c     &              P_data_leaving_South(ii, )
c            print*,'P_data_leaving_South(ii,10) = tau_xz_uno_rho2(i) ',
c     &              P_data_leaving_South(ii,10)
c            print*,'P_data_leaving_South(ii,  ) = tau_yy_uno_rho2(i) ',
c     &              P_data_leaving_South(ii,  )
c            print*,'P_data_leaving_South(ii,  ) = tau_yz_uno_rho2(i) ',
c     &              P_data_leaving_South(ii,  )
c            print*,'P_data_leaving_South(ii,11) = tau_zz_uno_rho2(i) ',
c     &              P_data_leaving_South(ii,11)
c            print*,'dudx_CSPH(i) ',dudx_CSPH(i)
cc           print*,'dudy_CSPH(i) ',dudy_CSPH(i)
c            print*,'dudz_CSPH(i) ',dudz_CSPH(i)
cc           print*,'dvdx_CSPH(i) ',dvdx_CSPH(i)
cc           print*,'dvdy_CSPH(i) ',dvdy_CSPH(i)
cc           print*,'dvdz_CSPH(i) ',dvdz_CSPH(i)
c            print*,'dwdx_CSPH(i) ',dwdx_CSPH(i)
cc           print*,'dwdy_CSPH(i) ',dwdy_CSPH(i)
c            print*,'dwdz_CSPH(i) ',dwdz_CSPH(i)
c            print*,'udot(i) ',udot(i)
cc           print*,'vdot(i) ',vdot(i)
c            print*,'wdot(i) ',wdot(i)
c            print*,'rdot(i) ',rdot(i)
c            !stop
c          endif
c        endif
c        !------------------------
      
      end do
      !--------------------------------------------------

      !call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point

      !--- Transfer Particles to neighbouring Processes ---

      !-- SendRecv particle numbers --
      !Send to North neighbour
c      call MPI_SENDRECV(
c     &  nb_leaving_North  , 1, MPI_INTEGER, north, 1,
c     &  nb_entering_South , 1, MPI_INTEGER, south, 1,
c     &  grid_comm, status, ierr)
c      call MPI_SENDRECV(
c     &  np_leaving_North  , 1, MPI_INTEGER, north, 1,
c     &  np_entering_South , 1, MPI_INTEGER, south, 1,
c     &  grid_comm, status, ierr)
c      call MPI_SENDRECV(
c     &  no_of_bfPs_leaving_North  , 1, MPI_INTEGER, north, 1,
c     &  no_of_bfPs_entering_South , 1, MPI_INTEGER, south, 1,
c     &  grid_comm, status, ierr)
c      call MPI_SENDRECV(
c     &  no_of_forcedmBPs_leaving_North  , 1, MPI_INTEGER, north, 1,
c     &  no_of_forcedmBPs_entering_South , 1, MPI_INTEGER, south, 1,
c     &  grid_comm, status, ierr)
c      call MPI_SENDRECV(
c     &  no_of_freemBPs_leaving_North  , 1, MPI_INTEGER, north, 1,
c     &  no_of_freemBPs_entering_South , 1, MPI_INTEGER, south, 1,
c     &  grid_comm, status, ierr)
c      !Send to South Neighbour
c      call MPI_SENDRECV(
c     &  nb_leaving_South  , 1, MPI_INTEGER, south, 0,
c     &  nb_entering_North , 1, MPI_INTEGER, north, 0,
c     &  grid_comm, status, ierr)
c      call MPI_SENDRECV(
c     &  np_leaving_South  , 1, MPI_INTEGER, south, 0,
c     &  np_entering_North , 1, MPI_INTEGER, north, 0,
c     &  grid_comm, status, ierr)
c      call MPI_SENDRECV(
c     &  no_of_bfPs_leaving_South  , 1, MPI_INTEGER, south, 0,
c     &  no_of_bfPs_entering_North , 1, MPI_INTEGER, north, 0,
c     &  grid_comm, status, ierr)
c      call MPI_SENDRECV(
c     &  no_of_forcedmBPs_leaving_South  , 1, MPI_INTEGER, south, 0,
c     &  no_of_forcedmBPs_entering_North , 1, MPI_INTEGER, north, 0,
c     &  grid_comm, status, ierr)
c      call MPI_SENDRECV(
c     &  no_of_freemBPs_leaving_South  , 1, MPI_INTEGER, south, 0,
c     &  no_of_freemBPs_entering_North , 1, MPI_INTEGER, north, 0,
c     &  grid_comm, status, ierr)
      np_nums_leaving_North  = 0
      np_nums_leaving_South  = 0  
      np_nums_entering_South = 0
      np_nums_entering_North = 0
      !- Pack North leaving particle numbers -
      np_nums_leaving_North(1) = nb_leaving_North
      np_nums_leaving_North(2) = np_leaving_North
      np_nums_leaving_North(3) = no_of_bfPs_leaving_North
      np_nums_leaving_North(4) = no_of_forcedmBPs_leaving_North
      np_nums_leaving_North(5) = no_of_freemBPs_leaving_North
      !Send to North Neighbour
      call MPI_SENDRECV(
     &    np_nums_leaving_North   , 1, iInt_IntegerType5, north, 107,
     &    np_nums_entering_South  , 1, iInt_IntegerType5, south, 107,
     &    grid_comm, status, ierr)
      !- UNPACK entering particle numbers -
      nb_entering_South               = np_nums_entering_South(1)
      np_entering_South               = np_nums_entering_South(2)
      no_of_bfPs_entering_South        = np_nums_entering_South(3)
      no_of_forcedmBPs_entering_South = np_nums_entering_South(4)
      no_of_freemBPs_entering_South   = np_nums_entering_South(5)
      !- Pack North leaving particle numbers -
      np_nums_leaving_South(1) = nb_leaving_South
      np_nums_leaving_South(2) = np_leaving_South
      np_nums_leaving_South(3) = no_of_bfPs_leaving_South
      np_nums_leaving_South(4) = no_of_forcedmBPs_leaving_South
      np_nums_leaving_South(5) = no_of_freemBPs_leaving_South
      !Send to South Neighbour
      call MPI_SENDRECV(
     &    np_nums_leaving_South   , 1, iInt_IntegerType5, south, 107,
     &    np_nums_entering_North  , 1, iInt_IntegerType5, north, 107,
     &    grid_comm, status, ierr)
      !- UNPACK entering particle numbers -
      nb_entering_North               = np_nums_entering_North(1)
      np_entering_North               = np_nums_entering_North(2)
      no_of_bfPs_entering_North        = np_nums_entering_North(3)
      no_of_forcedmBPs_entering_North = np_nums_entering_North(4)
      no_of_freemBPs_entering_North   = np_nums_entering_North(5)

      !call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point

!       if(rank_id.ge.i_rank_check.and.itime.ge.itime_check)then
!         print*
!         print*,'Sending Data index_check_NorthSouth.f'
!         print*,'itime, rank_id, north, south ',itime,rank_id,north,south
!         print*
!       end if

      !-- SendRecv Particle Data --
      !Send to North neighbour
      call MPI_SENDRECV(
     &  i_ABP_leaving_North  , nb_leaving_North , MPI_INTEGER, north, 1,
     &  i_ABP_entering_South , nb_entering_South, MPI_INTEGER, south, 1,
     &  grid_comm, status, ierr)
      !---Combined SendRecv's to North Neighbour ---
      call MPI_SENDRECV(
     &  P_xyz_data_leaving_North     , 1 , iP_xyzDataTypeB, north, 240,
     &  P_xyz_data_entering_South    , 1 , iP_xyzDataTypeB, south, 240,
     &  grid_comm, status, ierr)
      call MPI_SENDRECV(
     &  P_data_leaving_North     , 1 , iP_DataTypeIC , north, 241,
     &  P_data_entering_South    , 1 , iP_DataTypeIC , south, 241,
     &  grid_comm, status, ierr)
      if(i_filterTransfer_Perform.eq.1)then
       call MPI_SENDRECV(
     &  rhop_sum_data_leaving_North ,np_leaving_North ,MPI_REAL,north,1,
     &  rhop_sum_data_entering_South,np_entering_South,MPI_REAL,south,1,
     &  grid_comm, status, ierr)
      endif
      !- To be used with Verlet algorithm -
      if(i_algorithm.eq.2) then 
        call MPI_SENDRECV(
     &   P_verlet_data_leaving_North     , 1 , iP_DataTypeV, north, 242,
     &   P_verlet_data_entering_South    , 1 , iP_DataTypeV, south, 242,
     &   grid_comm, status, ierr)
      endif

      !end if   !End of:   if(np_leaving_north.gt.0.or.np_entering_South.gt.0)then

      !Send to South Neighbour
      call MPI_SENDRECV(
     &  i_ABP_leaving_South  , nb_leaving_South , MPI_INTEGER, south, 0,
     &  i_ABP_entering_North , nb_entering_North, MPI_INTEGER, north, 0,
     &  grid_comm, status, ierr)
      !---Combined SendRecv's to South Neighbour ---
      call MPI_SENDRECV(
     &  P_xyz_data_leaving_South     , 1 , iP_xyzDataTypeB, south, 243,
     &  P_xyz_data_entering_North    , 1 , iP_xyzDataTypeB, north, 243,
     &  grid_comm, status, ierr)
      call MPI_SENDRECV(
     &  P_data_leaving_South     , 1 , iP_DataTypeIC , south, 244,
     &  P_data_entering_North    , 1 , iP_DataTypeIC , north, 244,
     &  grid_comm, status, ierr)
      if(i_filterTransfer_Perform.eq.1)then
       call MPI_SENDRECV(
     &  rhop_sum_data_leaving_South ,np_leaving_South ,MPI_REAL,south,0,
     &  rhop_sum_data_entering_North,np_entering_North,MPI_REAL,north,0,
     &  grid_comm, status, ierr)
      endif
      !- To be used with Verlet algorithm -
      if(i_algorithm.eq.2) then 
        call MPI_SENDRECV(
     &   P_verlet_data_leaving_South     , 1 , iP_DataTypeV, south, 245,
     &   P_verlet_data_entering_North    , 1 , iP_DataTypeV, north, 245,
     &   grid_comm, status, ierr)
      endif
      !end if   !End of:  if(np_leaving_South.gt.0.or.np_entering_North.gt.0)then
c     -----------------------------------------

      !call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point

c     --- Prepare Data arrays for Unpacking ---
      net_nb_change = nb_entering_North + nb_entering_South
     &              - nb_leaving_North  - nb_leaving_South
      net_np_change = np_entering_North + np_entering_South
     &              - np_leaving_North  - np_leaving_South
      net_nbf_change = 
     &     no_of_bfPs_entering_North + no_of_bfPs_entering_South
     &   - no_of_bfPs_leaving_North  - no_of_bfPs_leaving_South
      net_nbfm_change =  
     &  no_of_forcedmBPs_entering_North +no_of_forcedmBPs_entering_South
     & -no_of_forcedmBPs_leaving_North  -no_of_forcedmBPs_leaving_South
      net_freemBP_change =  
     &   no_of_freemBPs_entering_North + no_of_freemBPs_entering_South
     & - no_of_freemBPs_leaving_North  - no_of_freemBPs_leaving_South

      !nptemp = np_local - no_of_Ps_leaving
      nbtemp = nb_local - nb_leaving_North - nb_leaving_South
      nptemp = np_local - np_leaving_North - np_leaving_South

      nbf_temp = nbf_local
     & - no_of_bfPs_leaving_North       - no_of_bfPs_leaving_South
      nbfm_temp = nbfm_local
     & - no_of_forcedmBPs_leaving_North - no_of_forcedmBPs_leaving_South
      nbfreem_temp = nbfreem_local
     & - no_of_freemBPs_leaving_North   - no_of_freemBPs_leaving_South
      nFP_temp = np_local - nb_local
     & - no_of_FPs_leaving_North        - no_of_FPs_leaving_South
      
      nbf_local     = nbf_local  + net_nbf_change
      nbfm_local    = nbfm_local + net_nbf_change
     &                           + net_nbfm_change
      nbfreem_local = nbfreem_local + net_nbf_change
     &                              + net_nbfm_change
     &                              + net_freemBP_change 
c      nb_local        = nb_local  + net_nbf_change
c     &                            + net_nbfm_change
c     &                            + net_freemBP_change
c      nbfp1_local = nbf_local + 1
c      nbp1_local  = nb_local  + 1

      if(net_nbf_change.ne.0.and.i_MPI_Adapt.ne.1)then
        print*
        print*,'Error in index_check.f'
        print*,'net_nbf_change.ne.0'
        print*,'itime ',itime
        print*,'rank_id        ',rank_id
        print*,'nbf_local      ',nbf_local
        print*,'nb_local       ',nb_local
        print*,'np_local       ',np_local
        print*,'net_nbf_change ',net_nbf_change
        stop
      endif

      nb_local_temp = nb_local + net_nb_change
      np_local_temp = np_local + net_np_change
      
c      !-- Debug Code------
c      if(itime.ge.itime_check)then
c        do i=0,n_procs-1
c          if(rank_id.eq.i)then
c            print*
c            print*,'rank_id ',rank_id
c            print*,'BPs leaving & entering'
c            print*,'nb_leaving_North  ',nb_leaving_North
c            print*,'nb_leaving_South  ',nb_leaving_South
c            print*,'nb_entering_North ',nb_entering_North
c            print*,'nb_entering_South ',nb_entering_South
c            print*,'Ps leaving & entering'
c            print*,'np_leaving_North  ',np_leaving_North
c            print*,'np_leaving_South  ',np_leaving_South
c            print*,'np_entering_North ',np_entering_North
c            print*,'np_entering_South ',np_entering_South
c            print*,'nbtemp = nb_local - nb_leaving ',nbtemp
c            print*,'nptemp = np_local - np_leaving ',nptemp
c            print*,'net_nb_change ',net_nb_change
c            print*,'net_np_change ',net_np_change
c            print*,'nb_local_temp ',nb_local_temp
c            print*,'net_nbf_change        ',net_nbf_change
c            print*,'net_nbfm_change  ',net_nbfm_change
c            print*,'net_freemBP_change       ',net_freemBP_change
c            print*,'nbf_local        ',nbf_local
c            print*,'nbfm_local  ',nbfm_local
c            print*,'nbfreem_local      ',nbfreem_local
c            print*,'nb_local         ',nb_local
c            print*,'np_local         ',np_local
c          end if
c          call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point
c        end do
c      end if

c     ***  fill up free spaces with new particles  ***

c     --- UNPACK DATA ENTERING WEST ---
c      !-- Debug Code------
c      if(rank_id.eq.i_rank_check.and.itime.ge.itime_check)then
c        print*
c        print*,'UnPacking Particle Data entering South'
c        print*,'rank_id ',rank_id
c        print*,'nb_entering_South ',nb_entering_South
c        print*,'np_entering_South ',np_entering_South
c        print*,'nb_local_temp    ',nb_local_temp
c        print*,'nbf_local          ',nbf_local
c        print*,'nbfm_local    ',nbfm_local
c        print*,'nbfreem_local        ',nbfreem_local
c        print*,'nbtemp          ',nbtemp
c        print*,'nptemp          ',nptemp
c        print*,'nbf_temp        ',nbf_temp
c        print*,'nbfm_temp  ',nbfm_temp
c        print*,'nbfreem_temp      ',nbfreem_temp
c      end if
c      !------------------------

      nbf_count       = 0
      nbfm_count      = 0
      nbfreem_count   = 0
      nFP_count       = 0
      
      do i = 1,np_entering_South

        nptemp = nptemp +1

        if(i.le.nb_entering_South)then
          if(iBPs_top_free_index .gt. 0)then
            ifree_index_ptr = iBPs_free_index(iBPs_top_free_index)
            iBPs_top_free_index = iBPs_top_free_index -1
          else if(iFPs_top_free_index .gt. 0)then
            ifree_index_ptr = iFPs_free_index(iFPs_top_free_index)
            iFPs_top_free_index = iFPs_top_free_index -1
          else!There are no free spaces in array other than right at the end
            !ifree_index_ptr = nptemp
            ifree_index_ptr = ihighest_np + 1
            ihighest_np     = ihighest_np + 1
          end if
          nbtemp = nbtemp +1
          ip_index_BP_local(nbtemp) = ifree_index_ptr
          !- Indexing -
          if(nbf_count.lt.no_of_bfPs_entering_South)then   !if(nbf_temp.lt.nbf_local)then
            nbf_count = nbf_count + 1   !nbf_temp = nbf_temp + 1
            nbf_index = nbf_count + nbf_temp
            i_particleType(ifree_index_ptr) = 1   !Fixed Boundary Particle
            ip_index_BP_local_1(nbf_index) = ifree_index_ptr
          else if(nbfm_count.lt.no_of_forcedmBPs_entering_South)then
            nbfm_count = nbfm_count + 1
            nbfm_index = nbfm_count + nbfm_temp
            i_particleType(ifree_index_ptr) = 2   !Forced Moving Boundary Particle
            ip_index_BP_local_2(nbfm_index) = ifree_index_ptr
          else if(nbfreem_count.lt.no_of_freemBPs_entering_South)then
            nbfreem_count = nbfreem_count + 1
            nbfreem_index = nbfreem_count + nbfreem_temp
            i_particleType(ifree_index_ptr) = 3   !Free Moving Boundary Particle
            ip_index_BP_local_3(nbfreem_index) = ifree_index_ptr
          end if
        else
          if(iFPs_top_free_index .gt. 0)then
            ifree_index_ptr = iFPs_free_index(iFPs_top_free_index)
            iFPs_top_free_index = iFPs_top_free_index -1
          else!There are no free spaces in array other than right at the end
            !ifree_index_ptr = nptemp
            ifree_index_ptr = ihighest_np + 1
            ihighest_np     = ihighest_np + 1
          end if
          nFP_count = nFP_count + 1
          nFP_index = nFP_count + nFP_temp
          !-- Error Checking Code------
          if(nFP_index.lt.1)then
            print*
            print*,'ERROR in index_check_NorthSouth.f #1'
            print*,'nFP_index.lt.1'
            print*,'itime, rank_id ',itime, rank_id
            print*,'rank_idx/y/z ',rank_idx, rank_idy, rank_idz
            stop
          end if
          ip_index_FP_local(nFP_index) = ifree_index_ptr
          !ip_index_FP_local(nptemp-nb_local_temp) = ifree_index_ptr
          i_particleType(ifree_index_ptr) = 4   !Fluid Particle
        end if

        if(iarray_loc_in_use(ifree_index_ptr).ne.0)then
          print*
          print*,'ERROR in Index_check.f: particles entering South'
          print*,'iarray_loc_in_use(ifree_index_ptr).ne.0'
          print*,'itime ',itime
          print*,'rank_id ',rank_id
          print*,'nb_local ',nb_local
          print*,'np_local ',np_local
          print*,'nb_local ',nb_local
          print*,'nbtemp  ',nbtemp
          print*,'nptemp  ',nptemp
          print*,'iBPs_top_free_index ',iBPs_top_free_index
          print*,'iFPs_top_free_index ',iFPs_top_free_index
          print*,'ifree_index_ptr ',ifree_index_ptr
          print*,'iarray_loc_in_use(ifree_index_ptr)',
     &            iarray_loc_in_use(ifree_index_ptr)
          print*,'BPs leaving & entering'
          print*,'nb_leaving_North  ',nb_leaving_North
          print*,'nb_leaving_South  ',nb_leaving_South
          print*,'nb_entering_North ',nb_entering_North
          print*,'nb_entering_South ',nb_entering_South
          print*,'Ps leaving & entering'
          print*,'np_leaving_North  ',np_leaving_North
          print*,'np_leaving_South  ',np_leaving_South
          print*,'np_entering_North ',np_entering_North
          print*,'np_entering_South ',np_entering_South
          print*,'nbtemp = nb_local - nb_leaving ',nbtemp
          print*,'nptemp = np_local - np_leaving ',nptemp
          print*,'net_nb_change ',net_nb_change
          print*,'net_np_change ',net_np_change
          print*,'nb_local_temp ',nb_local_temp
          print*,'net_nbf_change        ',net_nbf_change
          print*,'net_nbfm_change  ',net_nbfm_change
          print*,'net_freemBP_change       ',net_freemBP_change
          print*,'nbf_local        ',nbf_local
          print*,'nbfm_local  ',nbf_local
          print*,'nbfreem_local      ',nbf_local
          print*,'nb_local         ',nb_local
          stop     
        end if
        
        xp(ifree_index_ptr)             = P_xyz_data_entering_South(i,1)
        yp(ifree_index_ptr)             = P_xyz_data_entering_South(i,2)
        zp(ifree_index_ptr)             = P_xyz_data_entering_South(i,3)
        up(ifree_index_ptr)              = P_data_entering_South(i,1)
        vp(ifree_index_ptr)              = P_data_entering_South(i,2)
        wp(ifree_index_ptr)              = P_data_entering_South(i,3)
        rhop(ifree_index_ptr)            = P_data_entering_South(i,4)
        p(ifree_index_ptr)               = P_data_entering_South(i,5)
        pm(ifree_index_ptr)              = P_data_entering_South(i,6)
        pVol(ifree_index_ptr)            = P_data_entering_South(i,7)
        cs(ifree_index_ptr)              = P_data_entering_South(i,8)
        sum_wab(ifree_index_ptr)         = P_data_entering_South(i,9)
        tau_xx_uno_rho2(ifree_index_ptr) = P_data_entering_South(i,10) 
        tau_xy_uno_rho2(ifree_index_ptr) = P_data_entering_South(i,11)
        tau_xz_uno_rho2(ifree_index_ptr) = P_data_entering_South(i,12)
        tau_yy_uno_rho2(ifree_index_ptr) = P_data_entering_South(i,13) 
        tau_yz_uno_rho2(ifree_index_ptr) = P_data_entering_South(i,14)
        tau_zz_uno_rho2(ifree_index_ptr) = P_data_entering_South(i,15)
        TEp(ifree_index_ptr)             = P_data_entering_South(i,16)
        dudx_CSPH(ifree_index_ptr)       = P_data_entering_South(i,17)
        dudy_CSPH(ifree_index_ptr)       = P_data_entering_South(i,18)
        dudz_CSPH(ifree_index_ptr)       = P_data_entering_South(i,19)
        dvdx_CSPH(ifree_index_ptr)       = P_data_entering_South(i,20)
        dvdy_CSPH(ifree_index_ptr)       = P_data_entering_South(i,21)
        dvdz_CSPH(ifree_index_ptr)       = P_data_entering_South(i,22)
        dwdx_CSPH(ifree_index_ptr)       = P_data_entering_South(i,23)
        dwdy_CSPH(ifree_index_ptr)       = P_data_entering_South(i,24)
        dwdz_CSPH(ifree_index_ptr)       = P_data_entering_South(i,25)
        drhodx_CSPH(ifree_index_ptr)     = P_data_entering_South(i,26)
        drhody_CSPH(ifree_index_ptr)     = P_data_entering_South(i,27)
        drhodz_CSPH(ifree_index_ptr)     = P_data_entering_South(i,28)
        dTEdx_CSPH(ifree_index_ptr)      = P_data_entering_South(i,29)
        dTEdy_CSPH(ifree_index_ptr)      = P_data_entering_South(i,30)
        dTEdz_CSPH(ifree_index_ptr)      = P_data_entering_South(i,31)
        iarray_loc_in_use(ifree_index_ptr) = 1
        iflag(ifree_index_ptr) = 1

        if(i_filterTransfer_Perform.eq.1)then
          rhop_sum(ifree_index_ptr) = rhop_sum_data_entering_South(i)
        endif
        if(i.le.nb_entering_South)then
          iabsolute_BP_Pointer_Index(ifree_index_ptr) = 
     &                           i_ABP_entering_South(i)
        end if
        !- To be used with Verlet algorithm -
        if(i_algorithm.eq.2) then 
          uo(ifree_index_ptr)   = P_verlet_data_entering_South(i,1)
          vo(ifree_index_ptr)   = P_verlet_data_entering_South(i,2)
          wo(ifree_index_ptr)   = P_verlet_data_entering_South(i,3)
          rhoo(ifree_index_ptr) = P_verlet_data_entering_South(i,4)
          TEo(ifree_index_ptr)  = P_verlet_data_entering_South(i,5)
        endif

        !-- MPI Periodic OB correction --
!         !-- Periodic OB correction X-Direction --
!         if(iCartTop_periodicity(1))then
!          if(rank_idx.eq.0.and.xp(ifree_index_ptr).gt.xmax_container)then
!            dxp_dble = dble(xp(ifree_index_ptr)) - xmax_container_double    
!            xp(ifree_index_ptr) = real(dxp_dble +xmin_container_double)
!          else if(rank_idx.eq.n_procs_x-1.and.
!      &                       xp(ifree_index_ptr).lt.xmin_container)then
!            dxp_dble = xmin_container_double - dble(xp(ifree_index_ptr))   
!            xp(ifree_index_ptr) = real(xmax_container_double - dxp_dble)
!          end if
!         end if
        !-- Periodic OB correction Y-Direction --
        if(iCartTop_periodicity(2))then
         if(rank_idy.eq.0.and.yp(ifree_index_ptr).gt.ymax_container)then
           dyp_dble = dble(yp(ifree_index_ptr)) - ymax_container_double    
           yp(ifree_index_ptr) = real(dyp_dble +ymin_container_double)
         else if(rank_idy.eq.n_procs_y-1.and.
     &                       yp(ifree_index_ptr).lt.ymin_container)then
           dyp_dble = ymin_container_double - dble(yp(ifree_index_ptr))   
           yp(ifree_index_ptr) = real(ymax_container_double - dyp_dble)
         end if
        end if
!         !-- Periodic OB correction Z-Direction --
!         if(iCartTop_periodicity(3))then
!          if(rank_idz.eq.0.and.zp(ifree_index_ptr).gt.zmax_container)then
!            dzp_dble = dble(zp(ifree_index_ptr)) - zmax_container_double    
!            zp(ifree_index_ptr) = real(dzp_dble +zmin_container_double)
!          else if(rank_idz.eq.n_procs_z-1.and.
!      &                       zp(ifree_index_ptr).lt.zmin_container)then
!            dzp_dble = zmin_container_double - dble(zp(ifree_index_ptr))   
!            zp(ifree_index_ptr) = real(zmax_container_double - dzp_dble)
!          end if
!         end if

c        !-- Debug Code------
c        if(rank_id.eq.i_rank_check.and.itime.ge.itime_check)then
c          print*
c          print*,'i, ifree_index_ptr',
c     &            i, ifree_index_ptr
c          print*,'iBPs_top_free_index, iFPs_top_free_index',
c     &            iBPs_top_free_index, iFPs_top_free_index
c          print*,'xp, yp, zp(ifree_index_ptr) ',
c     &    xp(ifree_index_ptr), yp(ifree_index_ptr), zp(ifree_index_ptr) 
c          print*,'i_particleType(ifree_index_ptr) ',
c     &            i_particleType(ifree_index_ptr)
c          if(i.le.nb_entering_South)then
c            print*,'iabsolute_BP_Pointer_Index(ifree_index_ptr) ',
c     &              iabsolute_BP_Pointer_Index(ifree_index_ptr) 
c            print*,'nbf_count, nbfm_count, nbfreem_count ',
c     &              nbf_count, nbfm_count, nbfreem_count       
c            print*,'nbf_index, nbfm_index, nbfreem_index ',
c     &              nbf_index, nbfm_index, nbfreem_index       
c          end if
c        end if
c        !-------------------


      end do

c     --- UNPACK DATA ENTERING EAST ---
      !call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point
c      !-- Debug Code------
c      if(itime.ge.itime_check.and.rank_id.eq.i_rank_check)then
c        print*
c        print*,'UnPacking Particle Data entering North'
c        print*,'rank_id ',rank_id
c        print*,'nb_entering_North ',nb_entering_North
c        print*,'np_entering_North ',np_entering_North
c        print*,'nb_local_temp    ',nb_local_temp
c        print*,'nbf_local        ',nbf_local
c        print*,'nbfm_local  ',nbfm_local
c        print*,'nbfreem_local      ',nbfreem_local
c        print*,'nbtemp          ',nbtemp
c        print*,'nptemp          ',nptemp
c        print*,'nbf_temp        ',nbf_temp
c        print*,'nbfm_temp  ',nbfm_temp
c        print*,'nbfreem_temp      ',nbfreem_temp
c      end if
c      !------------------

      nbf_temp       = nbf_temp     + nbf_count
      nbfm_temp      = nbfm_temp    + nbfm_count
      nbfreem_temp   = nbfreem_temp + nbfreem_count
      nFP_temp       = nFP_temp     + nFP_count
      nbf_count      = 0
      nbfm_count     = 0
      nbfreem_count  = 0
      nFP_count      = 0
      
      do i = 1,np_entering_North

        nptemp = nptemp +1

        if(i.le.nb_entering_North)then
          if(iBPs_top_free_index .gt. 0)then
            ifree_index_ptr = iBPs_free_index(iBPs_top_free_index)
            iBPs_top_free_index = iBPs_top_free_index -1
          else if(iFPs_top_free_index .gt. 0)then
            ifree_index_ptr = iFPs_free_index(iFPs_top_free_index)
            iFPs_top_free_index = iFPs_top_free_index -1
          else!There are no free spaces in array other than right at the end
            !ifree_index_ptr = nptemp
            ifree_index_ptr = ihighest_np + 1
            ihighest_np     = ihighest_np + 1
          end if
          nbtemp = nbtemp +1
          ip_index_BP_local(nbtemp) = ifree_index_ptr
          !- Indexing -
          if(nbf_count.lt.no_of_bfPs_entering_North)then   !if(nbf_temp.lt.nbf_local)then
            nbf_count = nbf_count + 1   !nbf_temp = nbf_temp + 1
            nbf_index = nbf_count + nbf_temp
            i_particleType(ifree_index_ptr) = 1   !Fixed Boundary Particle
            ip_index_BP_local_1(nbf_index) = ifree_index_ptr
          else if(nbfm_count.lt.no_of_forcedmBPs_entering_North)then
            nbfm_count = nbfm_count + 1
            nbfm_index = nbfm_count + nbfm_temp
            i_particleType(ifree_index_ptr) = 2   !Forced Moving Boundary Particle
            ip_index_BP_local_2(nbfm_index) = ifree_index_ptr
          else if(nbfreem_count.lt.no_of_freemBPs_entering_North)then
            nbfreem_count = nbfreem_count + 1
            nbfreem_index = nbfreem_count + nbfreem_temp
            i_particleType(ifree_index_ptr) = 3   !Free Moving Boundary Particle
            ip_index_BP_local_3(nbfreem_index) = ifree_index_ptr
          end if
        else
          if(iFPs_top_free_index .gt. 0)then
            ifree_index_ptr = iFPs_free_index(iFPs_top_free_index)
            iFPs_top_free_index = iFPs_top_free_index -1
          else  !There are no free spaces in array other than right at the end
            !ifree_index_ptr = nptemp
            ifree_index_ptr = ihighest_np + 1
            ihighest_np     = ihighest_np + 1
          end if
          nFP_count = nFP_count + 1
          nFP_index = nFP_count + nFP_temp
          !-- Error Checking Code------
          if(nFP_index.lt.1)then
            print*
            print*,'ERROR in index_check_NorthSouth.f #2'
            print*,'nFP_index.lt.1'
            print*,'itime, rank_id ',itime, rank_id
            print*,'rank_idx/y/z ',rank_idx, rank_idy, rank_idz
            stop
          end if
          ip_index_FP_local(nFP_index) = ifree_index_ptr
          !ip_index_FP_local(nptemp-nb_local_temp) = ifree_index_ptr
          i_particleType(ifree_index_ptr) = 4   !Fluid Particle
        end if
        
        if(iarray_loc_in_use(ifree_index_ptr).ne.0)then
          print*
          print*,'ERROR in Index_check.f: particles entering North'
          print*,'iarray_loc_in_use(ifree_index_ptr).ne.0'
          print*,'itime ',itime
          print*,'rank_id ',rank_id
          print*,'nb_local ',nb_local
          print*,'np_local ',np_local
          print*,'nb_local ',nb_local
          print*,'nbtemp  ',nbtemp
          print*,'nptemp  ',nptemp
          print*,'ifree_index_ptr ',ifree_index_ptr
          print*,'iarray_loc_in_use(ifree_index_ptr)',
     &            iarray_loc_in_use(ifree_index_ptr)
          stop     
        end if

        xp(ifree_index_ptr)             = P_xyz_data_entering_North(i,1)
        yp(ifree_index_ptr)             = P_xyz_data_entering_North(i,2)
        zp(ifree_index_ptr)             = P_xyz_data_entering_North(i,3)
        up(ifree_index_ptr)              = P_data_entering_North(i,1)
        vp(ifree_index_ptr)              = P_data_entering_North(i,2)
        wp(ifree_index_ptr)              = P_data_entering_North(i,3)
        rhop(ifree_index_ptr)            = P_data_entering_North(i,4)
        p(ifree_index_ptr)               = P_data_entering_North(i,5)
        pm(ifree_index_ptr)              = P_data_entering_North(i,6)
        pVol(ifree_index_ptr)            = P_data_entering_North(i,7)
        cs(ifree_index_ptr)              = P_data_entering_North(i,8)
        sum_wab(ifree_index_ptr)         = P_data_entering_North(i,9)
        tau_xx_uno_rho2(ifree_index_ptr) = P_data_entering_North(i,10) 
        tau_xy_uno_rho2(ifree_index_ptr) = P_data_entering_North(i,11)
        tau_xz_uno_rho2(ifree_index_ptr) = P_data_entering_North(i,12)
        tau_yy_uno_rho2(ifree_index_ptr) = P_data_entering_North(i,13) 
        tau_yz_uno_rho2(ifree_index_ptr) = P_data_entering_North(i,14)
        tau_zz_uno_rho2(ifree_index_ptr) = P_data_entering_North(i,15)
        TEp(ifree_index_ptr)             = P_data_entering_North(i,16)
        dudx_CSPH(ifree_index_ptr)       = P_data_entering_North(i,17)
        dudy_CSPH(ifree_index_ptr)       = P_data_entering_North(i,18)
        dudz_CSPH(ifree_index_ptr)       = P_data_entering_North(i,19)
        dvdx_CSPH(ifree_index_ptr)       = P_data_entering_North(i,20)
        dvdy_CSPH(ifree_index_ptr)       = P_data_entering_North(i,21)
        dvdz_CSPH(ifree_index_ptr)       = P_data_entering_North(i,22)
        dwdx_CSPH(ifree_index_ptr)       = P_data_entering_North(i,23)
        dwdy_CSPH(ifree_index_ptr)       = P_data_entering_North(i,24)
        dwdz_CSPH(ifree_index_ptr)       = P_data_entering_North(i,25)
        drhodx_CSPH(ifree_index_ptr)     = P_data_entering_North(i,26)
        drhody_CSPH(ifree_index_ptr)     = P_data_entering_North(i,27)
        drhodz_CSPH(ifree_index_ptr)     = P_data_entering_North(i,28)
        dTEdx_CSPH(ifree_index_ptr)      = P_data_entering_North(i,29)
        dTEdy_CSPH(ifree_index_ptr)      = P_data_entering_North(i,30)
        dTEdz_CSPH(ifree_index_ptr)      = P_data_entering_North(i,31)
        iarray_loc_in_use(ifree_index_ptr) = 1
        iflag(ifree_index_ptr) = 1

        if(i_filterTransfer_Perform.eq.1)then
          rhop_sum(ifree_index_ptr) = rhop_sum_data_entering_North(i)
        endif
        if(i.le.nb_entering_North)then
          iabsolute_BP_Pointer_Index(ifree_index_ptr) = 
     &                           i_ABP_entering_North(i)
        end if
        !- To be used with Verlet algorithm -
        if(i_algorithm.eq.2) then 
          uo(ifree_index_ptr)   = P_verlet_data_entering_North(i,1)
          vo(ifree_index_ptr)   = P_verlet_data_entering_North(i,2)
          wo(ifree_index_ptr)   = P_verlet_data_entering_North(i,3)
          rhoo(ifree_index_ptr) = P_verlet_data_entering_North(i,4)
          TEo(ifree_index_ptr)  = P_verlet_data_entering_North(i,5)
        endif

        !-- MPI Periodic OB correction --
        !-- Periodic OB correction X-Direction --
!         if(iCartTop_periodicity(1))then
!          if(rank_idx.eq.0.and.xp(ifree_index_ptr).gt.xmax_container)then
!            dxp_dble = dble(xp(ifree_index_ptr)) - xmax_container_double    
!            xp(ifree_index_ptr) = real(dxp_dble +xmin_container_double)
!          else if(rank_idx.eq.n_procs_x-1.and.
!      &                       xp(ifree_index_ptr).lt.xmin_container)then
!            dxp_dble = xmin_container_double - dble(xp(ifree_index_ptr))   
!            xp(ifree_index_ptr) = real(xmax_container_double - dxp_dble)
!          end if
!         end if
        !-- Periodic OB correction Y-Direction --
        if(iCartTop_periodicity(2))then
         if(rank_idy.eq.0.and.yp(ifree_index_ptr).gt.ymax_container)then
           dyp_dble = dble(yp(ifree_index_ptr)) - ymax_container_double    
           yp(ifree_index_ptr) = real(dyp_dble +ymin_container_double)
         else if(rank_idy.eq.n_procs_y-1.and.
     &                       yp(ifree_index_ptr).lt.ymin_container)then
           dyp_dble = ymin_container_double - dble(yp(ifree_index_ptr))   
           yp(ifree_index_ptr) = real(ymax_container_double - dyp_dble)
         end if
        end if
!         !-- Periodic OB correction Z-Direction --
!         if(iCartTop_periodicity(3))then
!          if(rank_idz.eq.0.and.zp(ifree_index_ptr).gt.zmax_container)then
!            dzp_dble = dble(zp(ifree_index_ptr)) - zmax_container_double    
!            zp(ifree_index_ptr) = real(dzp_dble +zmin_container_double)
!          else if(rank_idz.eq.n_procs_z-1.and.
!      &                       zp(ifree_index_ptr).lt.zmin_container)then
!            dzp_dble = zmin_container_double - dble(zp(ifree_index_ptr))   
!            zp(ifree_index_ptr) = real(zmax_container_double - dzp_dble)
!          end if
!         end if

c        !-- Debug Code------
c        if(itime.ge.itime_check.and.rank_id.eq.i_rank_check)then
c        !if(itime.ge.itime_check.and.rank_id.eq.3)then
c          print*,'i, ifree_index_ptr',
c     &            i, ifree_index_ptr
c          print*,'iBPs_top_free_index, iFPs_top_free_index',
c     &            iBPs_top_free_index, iFPs_top_free_index
c          print*,'xp, yp, zp(ifree_index_ptr) ',
c     &    xp(ifree_index_ptr), yp(ifree_index_ptr), zp(ifree_index_ptr) 
cc          print*,'xp(ifree_index_ptr)   ',
cc     &            xp(ifree_index_ptr)
cc          print*,'zp(ifree_index_ptr)   ',
cc     &            zp(ifree_index_ptr)
cc          print*,'up(ifree_index_ptr)   ',
cc     &            up(ifree_index_ptr)
cc          print*,'wp(ifree_index_ptr)   ',
cc     &            wp(ifree_index_ptr)
cc          print*,'rhop(ifree_index_ptr) ',
cc     &            rhop(ifree_index_ptr)
cc          print*,'p(ifree_index_ptr)    ',
cc     &            p(ifree_index_ptr)
cc          print*,'pm(ifree_index_ptr)   ',
cc     &            pm(ifree_index_ptr)
cc          print*,'pVol(ifree_index_ptr) ',
cc     &            pVol(ifree_index_ptr)
cc          print*,'cs(ifree_index_ptr)   ',
cc     &            cs(ifree_index_ptr)
c          print*,'i_particleType(ifree_index_ptr) ',
c     &            i_particleType(ifree_index_ptr)
c         if(i.le.nb_entering_North)then
c            print*,'iabsolute_BP_Pointer_Index(ifree_index_ptr) ',
c     &              iabsolute_BP_Pointer_Index(ifree_index_ptr) 
c            print*,'nbf_count, nbfm_count, nbfreem_count ',
c     &              nbf_count, nbfm_count, nbfreem_count       
c            print*,'nbf_index, nbfm_index, nbfreem_index ',
c     &              nbf_index, nbfm_index, nbfreem_index       
c          end if
c        end if
c        !-----------------------

      end do
      
      if(nbtemp.ne.nb_local_temp)then
        print*
        print*,'ERROR in Index_check.f'
        print*,'itime ',itime
        print*,'rank_id ',rank_id
        print*,'nbtemp.ne.nb_local_temp'
        stop        
      end if
      if(nptemp.ne.np_local_temp)then
        print*
        print*,'ERROR in Index_check.f'
        print*,'itime ',itime
        print*,'rank_id ',rank_id
        print*,'nptemp.ne.np_local_temp'
        stop        
      end if

      nb_local = nbtemp
      np_local = nptemp
      nbfp1_local = nbf_local + 1
      nbp1_local  = nb_local  + 1
      if(iBC.eq.1)then
         nstart=nbp1_local
      else
         nstart=1
      endif
      nstart_minus1 = nstart - 1
      
      if(nptemp > ihighest_np) ihighest_np = np_local


!       !-- Debug Code------
!        call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point
!       if(itime.ge.itime_check.and.rank_id.eq.i_rank_check)then
!         print*
!         print*,'All Data Entering UnPacked'
!         print*,'itime ',itime
!         print*,'rank_id ',rank_id
!         print*,'nb_entering_North ',nb_entering_North
!         print*,'np_entering_North ',np_entering_North
!         print*,'nb_entering_South ',nb_entering_South
!         print*,'np_entering_South ',np_entering_South
!         print*,'nbf_local          ',nbf_local
!         print*,'nbfm_local    ',nbfm_local
!         print*,'nbfreem_local        ',nbfreem_local
!         print*,'nb_local           ',nb_local
!         print*,'np_local           ',np_local
!         print*
!         print*,'Performing Non-MPI Periodic Boundary Correction'
!         print*
!       end if
!       !--------------------

      !-- Periodic Open Boundary Correction --
      do kpcount=1,np_local
        if(kpcount.le.nb_local)then
          k = ip_index_BP_local(kpcount)
        else
          k = ip_index_FP_local(kpcount-nb_local)
        end if
        !-- Periodic OB correction X-Direction --
        if(i_periodicOBs(1).eq.1)then
          if(xp(k).gt.xmax_container)then
            dxp_dble = dble(xp(k)) - xmax_container_double 
            xp(k) = real(dxp_dble +xmin_container_double)
          else if(xp(k).lt.xmin_container)then
            dxp_dble = xmin_container_double - dble(xp(k))   
            xp(k) = real(xmax_container_double - dxp_dble)
          end if
        end if
        !-- Periodic OB correction Y-Direction --
        if(i_periodicOBs(2).eq.1)then
          if(yp(k).gt.ymax_container)then
            dyp_dble = dble(yp(k)) - ymax_container_double 
            yp(k) = real(dyp_dble +ymin_container_double)
          else if(yp(k).lt.ymin_container)then
            dyp_dble = ymin_container_double - dble(yp(k))   
            yp(k) = real(ymax_container_double - dyp_dble)
          end if
        end if
        !-- Periodic OB correction Z-Direction --
        if(i_periodicOBs(3).eq.1)then
          if(zp(k).gt.zmax_container)then
            dzp_dble = dble(zp(k)) - zmax_container_double 
            zp(k) = real(dzp_dble +zmin_container_double)
          else if(zp(k).lt.zmin_container)then
            dzp_dble = zmin_container_double - dble(zp(k))   
            zp(k) = real(zmax_container_double - dzp_dble)
          end if
        end if
      end do

c      !-- Debug Code------
c      if(itime.ge.itime_check.and.rank_id.eq.i_rank_check)then
c        print*
c        print*,'Non-MPI Periodic OB correction Complete'
c        print*,'itime ',itime
c        print*,'rank_id ',rank_id
c        print*
c        print*,'Checking Particle Index Pointers'
c        print*
c        print*
c      end if
c      !-------------------

      !ihighest_np = np_local  !only use this for periodic boundaries
c     ..

c     ***  reorder the ip_index  for the next loop  ***
      !This ensures that after particles have left,
      !the linked list when accessed sequentially
      !only looks at array locations that contain particles
      !-- Serial Version --
      !ip_index_no = nb
      !do k = nbp1,ihighest_np
      !   if(iarray_loc_in_use(k).eq.1)then
      !      ip_index_no = ip_index_no+1
      !      ip_index(ip_index_no) = k
      !   end if
      !end do
      !-- MPI Version --
      ip_index_no_BPs = 0
      ip_index_no_FPs = 0
      kk_highest = 0
      ip_index_no_BPs_1 = 0
      ip_index_no_BPs_2 = 0
      ip_index_no_BPs_3 = 0
      do k = 1,ihighest_np
        if(iarray_loc_in_use(k).eq.1)then
         if((yp(k).lt.ymin_rank.or.yp(k).gt.ymax_rank))then
          print*
          print*,'Problem with particle location in '
     &                //'index_check_Conservative_NorthSouth.f'
          print*,'itime ',itime
          print*,'rank_id ',rank_id
          print*,'rank_idx, rank_idy, rank_idz ',
     &            rank_idx, rank_idy, rank_idz
          print*,'nb_local ',nb_local
          print*,'np_local ',np_local
          print*,'ymin_rank ',ymin_rank
          print*,'ymax_rank ',ymax_rank
          print*,'k ',k
          print*,'xp(k), yp(k), zp(k) ',xp(k), yp(k), zp(k)
          print*,'i_particleType(k)    ',i_particleType(k)
          print*,'iarray_loc_in_use(k) ',iarray_loc_in_use(k)
          stop
         end if
          if(i_particleType(k).lt.4)then  !Boundary Particle
            ip_index_no_BPs = ip_index_no_BPs+1
            ip_index_BP_local(ip_index_no_BPs) = k
            if(i_particleType(k).eq.1)then
              ip_index_no_BPs_1 = ip_index_no_BPs_1 + 1
              ip_index_BP_local_1(ip_index_no_BPs_1) = k
            else if(i_particleType(k).eq.2)then
              ip_index_no_BPs_2 = ip_index_no_BPs_2 + 1
              ip_index_BP_local_2(ip_index_no_BPs_2) = k
            else if(i_particleType(k).eq.3)then
              ip_index_no_BPs_3 = ip_index_no_BPs_3 + 1
              ip_index_BP_local_3(ip_index_no_BPs_3) = k
            else  !Error catching
              print*,'Problem #1 with i_particleType(k) in '
     &               //'index_check_NorthSouth.f'
              print*,'itime ',itime
              print*,'rank_id ',rank_id
              print*,'k ',k
              print*,'xp(k), yp(k), zp(k) ',xp(k), yp(k), zp(k)
              print*,'i_particleType(k) ',i_particleType(k)
              stop
            end if
          else if(i_particleType(k).eq.4)then  !Fluid Particle
            ip_index_no_FPs = ip_index_no_FPs+1
            ip_index_FP_local(ip_index_no_FPs) = k
          else  !Error catching
              print*,'Problem #2 with i_particleType(k) in '
     &               //'index_check_NorthSouth.f'
            print*,'itime ',itime
            print*,'rank_id ',rank_id
            print*,'k ',k
            print*,'xp(k), yp(k), zp(k) ',xp(k), yp(k), zp(k)
            print*,'i_particleType(k) ',i_particleType(k)
            stop
          end if
          if(k.gt.kk_highest) kk_highest = k   !Used for starting the imported image particles count
        end if
      end do
      if(kk_highest.gt.ihighest_np)then
        print*
        print*,'ERROR in Index_check.f'
        print*,'itime ',itime
        print*,'rank_id ',rank_id
        print*,'kk_highest.gt.ihighest_np'
        print*,'kk_highest  ',kk_highest
        print*,'ihighest_np ',ihighest_np
        print*
        stop        
      endif
      kk_highest = ihighest_np + 1
      
c      !- Image Printout to Screen- 
c      if(itime.ge.itime_check.and.rank_id.eq.i_rank_check)then
c      !if(itime.ge.itime_check.and.rank_id.eq.2)then
c        print*
c        print*,'rank_id ',rank_id
c        print*,'itime   ',itime
c        print*,'nbf_local       ' ,nbf_local 
c        print*,'nbfm_local ' ,nbfm_local 
c        print*,'nbfreem_local     ' ,nbfreem_local 
c        print*,'nb_local       ',nb_local
c        print*,'np_local       ',np_local
c        do ipcount=1,np_local
c          if(ipcount.le.nbf_local)then
c            i = ip_index_BP_local_1(ipcount)
c          else if(ipcount.le.nbfm_local)then
c            i = ip_index_BP_local_2(ipcount-nbf_local)
c          else if(ipcount.le.nbfreem_local)then
c            i = ip_index_BP_local_3(ipcount-nbfm_local)
c          else
c            i = ip_index_FP_local(ipcount-nb_local)
c          end if
c          print*,'ipcount -> i, Type, xp, zp',
c     &                ipcount,i,i_particleType(i),xp(i),zp(i)
c        end do
c      endif
     
      if(i_MPI_Adapt.eq.1.and.n_procs_y.gt.1)then
        if(itime/n_itime_Adapt*n_itime_Adapt.eq.itime
     &                      .or.itime.eq.itime_init)then
          call Adapt_MPI_Regrid
        endif
      endif
      
      !-  Update indices & FLOATING BODY distances from Body Centre -
      if(num_FB.gt.0)then
        !- Clear indexing system
        !do i_num_FB = 1,num_FB
        !  nb_FB_local(i_num_FB) = nbfm_local
        !enddo
        do ipcount = nbfm_local+1,nbfreem_local
          i = ip_index_BP_local_3(ipcount - nbfm_local)
          i_absolute = iabsolute_BP_Pointer_Index(i)
          i_num_FB = i_FB_Pointer_Info(i_absolute)
          !- Error catching -
          if(i_num_FB.lt.1)then
            print*,'ERROR in index_check_3D.f'
            print*,'i_num_FB.lt.1'
            print*,'rank_id, itime ',rank_id, itime
            print*,'nbfm_local+1 ',nbfm_local+1
            print*,'nbfreem_local       ',nbfreem_local
            print*,'i              ',i
            print*,'i_absolute     ',i_absolute
            print*,'i_num_FB       ',i_num_FB
            stop
          endif
c          xp_temp = BP_xyz_Data(i_absolute,1)
c          yp_temp = BP_xyz_Data(i_absolute,2)
c          zp_temp = BP_xyz_Data(i_absolute,3)
c          xp_minus_R0(i) = xp_temp - Box_XC(i_num_FB)
c          yp_minus_R0(i) = yp_temp - Box_YC(i_num_FB)
c          zp_minus_R0(i) = zp_temp - Box_ZC(i_num_FB)
        enddo
      endif


c      !- Image Printout to Screen- 
c      if(itime.ge.itime_check.and.rank_id.eq.3)then
c      !if(itime.ge.itime_check.and.rank_id.eq.2)then
c        print*
c        print*,'rank_id ',rank_id
c        print*,'itime   ',itime
c        print*,'nb_local       ',nb_local
c        print*,'np_local       ',np_local
c        do ipcount=1,np_local
c          !if(ipcount.le.nb_local)then
c          !  i = ip_index_BP_local(ipcount)
c          !else
c          !  i = ip_index_FP_local(ipcount-nb_local)
c          !end if
c          if(ipcount.le.nbf_local)then
c            i = ip_index_BP_local_1(ipcount)
c          else if(ipcount.le.nbfm_local)then
c            i = ip_index_BP_local_2(ipcount-nbf_local)
c          else if(ipcount.le.nbfreem_local)then
c            i = ip_index_BP_local_3(ipcount-nbfm_local)
c          else
c            i = ip_index_FP_local(ipcount-nb_local)
c          end if
c          print*,'ipcount -> i, Type, xp, yp, zp',
c     &                ipcount,i,i_particleType(i),xp(i),yp(i),zp(i)
c        end do
c        print*,'kk_highest, ihighest_np',kk_highest, ihighest_np
c        print*,'ip_index_no_BPs ,ip_index_no_FPs ',
c     &          ip_index_no_BPs ,ip_index_no_FPs
c        no_array_positions = 0
c        !do ipcount = 1,nb_local
c        !  i = ip_index_BP_local(ipcount)
c        !enddo
c        do kk = 1,ihighest_np
c          print*,'kk, iarray_loc_in_use(kk), xp, yp, zp ', 
c     &            kk, iarray_loc_in_use(kk),xp(kk),yp(kk),zp(kk)
c          if(iarray_loc_in_use(kk).eq.1)then
c            no_array_positions = no_array_positions +1
c          endif
c        enddo
c        print*,'no_array_positions ',no_array_positions
c        print*,'nb_local        ',nb_local
c        print*,'np_local        ',np_local
c        print*,'iBPs_top_free_index ',iBPs_top_free_index
c        do i=1,iBPs_top_free_index
c          print*,'i,iBPs_free_index(i) ',i,iBPs_free_index(i)
c        enddo
c        print*,'iFPs_top_free_index ',iFPs_top_free_index
c        do i=1,iFPs_top_free_index
c          print*,'i,iFPs_free_index(i) ',i,iFPs_free_index(i)
c        enddo        
c        if(net_np_change.gt.0)then
c          print*,'ifree_index_ptr     ',ifree_index_ptr
c        endif
c        !stop
c      end if

c      !-- Debug Code------
c      if(itime.ge.itime_check)then
c        do i=0,n_procs-1
c          if(rank_id.eq.i)then   !if(i.eq.-rank_id-1)then   !if(i.eq.rank_id)then
c            print*
c            print*,'itime ',itime
c            write(*,*),'rank_id ', rank_id
c            print*,'nbf_local ',nbf_local
c            print*,'nb_local  ',nb_local
c            print*,'nb_local   ',nb_local
c            print*,'np_local   ',np_local
c          end if
c          call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point
c        end do
c      end if
c      if(itime.ge.itime_check.and.rank_id.eq.i_rank_check)then   !if(rank_id.eq.-10)then
c        print*
c        print*,'index_check finished'
c        print*,'itime ',itime
c        print*,'rank_id ',rank_id
c        !stop
c      end if
c      !---------------------

c     .. Inflow Periodic OB correction .. 
      if(iCartTop_periodicity(1).and.rank_id.eq.-10)then
        if(rank_id.eq.-10)then
          print*
          print*,'Performing Inflow Periodic OB correction'
        end if      
        n_inlet = 0
        flow_rate = 0.0
        flow_in_old_temp = flow_in_old
        do ipcount=nb_local+1,np_local
          i = ip_index_FP_local(ipcount-nb_local)
          if(xp(i).gt.xmin_rank.and.xp(i).lt.xmin_rank + 2.0*h)then
            n_inlet = n_inlet + 1
            flow_rate = flow_rate + up(i)           
c            print*,'n_inlet, xp(i), up(i), flow_rate',
c     &              n_inlet, xp(i), up(i), flow_rate
          end if
        end do
        flow_rate = flow_rate / n_inlet
        flow_force = (2.0*(flow_rate - U_in) - (flow_in_old - U_in))
     &             /(2.0*dt)
        grx = - flow_force
        if(itime.lt.1000)then
          grx = grx*real(itime)/real(1000)
        end if
        flow_in_old = flow_rate
c        print*,'rank_id,itime,U_in,flow_in_old,flow_rate,flow_force,
c     &n_inlet'
c        write(*,*)rank_id,itime,U_in,flow_in_old,flow_rate,
c     &            flow_force,n_inlet

        call MPI_BCast(grx, 1, MPI_REAL, 
     &               0, MPI_COMM_WORLD, ierr)
        if(rank_id.eq.-10)then
          !print*
          write(*,366),itime, rank_id, n_inlet,
     &      grx, U_in,flow_in_old_temp,flow_rate,flow_force
          print*
        end if
      end if
366   format(i6,i3,i5,6(' ',e12.6))


!       if(itime.ge.itime_check.and.rank_id.ge.i_rank_check)then
!         print*
!         print*,'END OF index_check_NorthSouth.f'
!         print*,'itime, rank_id ',itime, rank_id
!         print*,'np_local ',np_local
!         !if(itime.eq.0)stop
!       end if

      !stop
c
800   return
      end !subroutine
