c    "Copyright 2010 Prof. Robert Dalrymple, Prof. M. Gomez Gesteira, Dr Benedict Rogers, 
c     Dr Alejandro Crespo, Dr Muthukumar Narayanaswamy, Dr Shan Zou, Dr Andrea Panizzo "
c
c    This file is part of SPHYSICS.
c
c    SPHYSICS is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 3 of the License, or
c    (at your option) any later version.
c
c    SPHYSICS is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program.  If not, see <http://www.gnu.org/licenses/>.


      subroutine step

      include 'common.3D'
      


C	Verlet Method

c  ...  compute acceleration
c

c      !-- Debug Code------
c      itime_check = 10e8   !0    !10e8   !0    !
c      i_rank_check = 3
c      i_ParticleCheck = 2527          
      
      if(rank_id.eq.0.and.itime/10*10 .eq. itime) then
        write(*,*) 'In step:   time = ',time,' itime = ',itime
      endif

       nb_image_local      = nb_local
       np_image_local      = np_local
       nbf_image_local     = nbf_local
       nbfm_image_local    = nbfm_local
       nbfreem_image_local = nbfreem_local
       nbp1_local = nb_local + 1
       nfbp1_local = nbf_local + 1     

c     --- Primary Divide for Resident Fluid Particles only ---
       i_resident_0 = 1  !Resident Moving Particles
       call check_limits_3D      
       call ini_divide(2)
       call divide(nbp1_local,np_local,2,i_resident_0)   !3rd argument = kind_p)  

c      --- Primary Divide for Resident Moving Boundary Particles only ---
       !call recover_list   !Why here nad not below? (see step_predictor_corrector_MPI_2D.f)
       if(iopt_movingObject.eq.1)then
         call recover_list
         call divide(nbfp1_local,nb_local,1,i_resident_0)   !3rd argument = kind_p)
       endif

      call divide_ParallelPrep(1)   !argument 1 selects densityFilter option

      
      call ac          


c
c  ... compute corrections to:
c         (a) rate of change of velocity due to
c             (i)  body forces, and
c             (ii) boundary forces
c
c         (b) rate of change of position (free-surface, XSPH)
c
C
      call correct
	
	
c     To calculate time step 
      if (ivar_dt.eq.1) then
	    call variable_time_step  
	    dt = dt_new
      endif
	
      dtsq_05 = 0.5*dt*dt
      two_dt=2*dt
      dt2 = dt

C

      do ipcount=nstart,np_local
          call ipcount_to_i(ipcount,i)
          um1(i)  = uo(i)
          vm1(i)  = vo(i)
          wm1(i)  = wo(i)
          xo(i)   = xp(i)
          yo(i)   = yp(i)
          zo(i)   = zp(i)
          uo(i)   = up(i)
          vo(i)   = vp(i)
          wo(i)   = wp(i)
          rhom1(i)= rhoo(i)
          rhoo(i) = rhop(i)
          TEm1(i) = TEo(i)    
	   TEo(i)  = TEp(i)
      enddo

c
c     determine which velocity term to use

      nstep_verlet=nstep_verlet+1
      imix=0
      if(mod(nstep_verlet,40).eq.0)then
        imix=1     
        nstep_verlet = 0
      end if

      if (imix.eq.0) then
         do ipcount=nb_local+1,np_local
            call ipcount_to_i(ipcount,i)
            xp(i) = xo(i) +xdot(i)*dt +udot(i)*dtsq_05        
            yp(i) = yo(i) +ydot(i)*dt +vdot(i)*dtsq_05 
            zp(i) = zo(i) +zdot(i)*dt +wdot(i)*dtsq_05 
           
            up(i) = um1(i) + two_dt*udot(i)
            vp(i) = vm1(i) + two_dt*vdot(i)
            wp(i) = wm1(i) + two_dt*wdot(i)
          enddo
      else
          do ipcount=nb_local+1,np_local
            call ipcount_to_i(ipcount,i)
            xp(i) = xo(i) +xdot(i)*dt +udot(i)*dtsq_05          
            yp(i) = yo(i) +ydot(i)*dt +vdot(i)*dtsq_05 
            zp(i) = zo(i) +zdot(i)*dt +wdot(i)*dtsq_05 

            up(i) = uo(i)+udot(i)*dt
            vp(i) = vo(i)+vdot(i)*dt
            wp(i) = wo(i)+wdot(i)*dt
          enddo
      endif

      if(imix.eq.0) then
        do ipcount=nstart,np_local
           call ipcount_to_i(ipcount,i)
           rhop(i) = rhom1(i) + two_dt*rdot(i)
           TEp(i) = TEm1(i) + two_dt*TEdot(i) 
c    ...   compute pressure using the equation of state
           call equation_of_state(rhop(i),TEp(i),p(i),cs(i))
        enddo
      else
        do ipcount=nstart,np_local
           call ipcount_to_i(ipcount,i)
           rhop(i) = rhoo(i) + dt*rdot(i)
           TEp(i) = TEo(i) + dt*TEdot(i) 
c    ...   compute pressure using the equation of state
           call equation_of_state(rhop(i),TEp(i),p(i),cs(i))
        enddo
      endif


      call movingObjects(1)


      call index_check_EastWest
      if(i_call_UpDown    .eq.1) call index_check_UpDown
      if(i_call_NorthSouth.eq.1) call index_check_NorthSouth  !3-D only

c     Normals are calculated in subroutine updateNormals
      if(iBC.eq.1)then   !Repulsive Force BC
       i_particleTypeChoice = 0
       if(i_MPI_Adapt.eq.1)then
         nbf_uN_start = 1             !All BPs
       else
         nbf_uN_start = nbf_local+1   !All moving BPs
       endif
       call updateNormals_3D(nbf_uN_start,nb_local,i_particleTypeChoice)   
      end if

c      call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point

      end