c    "Copyright 2010 Prof. Robert Dalrymple, Prof. M. Gomez Gesteira, Dr Benedict Rogers, 
c     Dr Alejandro Crespo, Dr Muthukumar Narayanaswamy, Dr Shan Zou, Dr Andrea Panizzo "
c
c    This file is part of SPHYSICS.
c
c    SPHYSICS is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 3 of the License, or
c    (at your option) any later version.
c
c    SPHYSICS is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program.  If not, see <http://www.gnu.org/licenses/>.

      subroutine updateNormals_3D(i_start,i_stop,i_particleTypeChoice)

      include 'common.3D'

      double precision dxpb1,dxpb2
      
c       Subroutine calculates tangents and normals for a surface given 
c       by three points:
c
c       Vectors t, s, n form a righ-handed (orthogonal) system
c
c
c       (xi     , yi    , zi )    = point of interest 
c       (xt_next,yt_next,zt_next) = next point in traverse of plane
c       (xs_next,ys_next,zs_next) = corresponding point in orthogonal traverse

!               ^ (drs)
!              /
!            
!       ^(n) o(3)
!       |   /
!       |  /
!       | /
!       |/ )
!    (i)o---------o(2)  -->(drt)
c    
c      i_start = 1         !BedPs only
c      i_stop  = nbf       !Bed particles only
c
c      i_start = nbf + 1   !Moving Solid Object particles only
c      i_stop  = nb        !Moving Solid Object particles only
c
c       Specifically written for use in Monaghan's BC

    
      !-- MPI Version --
      do ipcount=i_start,i_stop  
       call ipcount_to_i(ipcount,i)
        
       if(i.lt.1)then
         print*
         print*,'ERROR in updateNormals_3D'
         print*,'rank_id, itime  ',rank_id, itime
         print*,'i_start, i_stop ',i_start,i_stop
         print*,'nbf_local       ' ,nbf_local 
         print*,'nbfm_local      ' ,nbfm_local 
         print*,'nbfreem_local   ' ,nbfreem_local 
         print*
         stop
       endif
         
        xp_i = xp(i)
        yp_i = yp(i)
        zp_i = zp(i)
        
        i_absolute = iabsolute_BP_Pointer_Index(i)
        iminus1_neighbour = iBP_Pointer_Info(i_absolute,3)
        iplus1_neighbour  = iBP_Pointer_Info(i_absolute,4)
        jminus1_neighbour = iBP_Pointer_Info(i_absolute,5)
        jplus1_neighbour  = iBP_Pointer_Info(i_absolute,6)
        
        xp_BP_iminus1 = BP_xyz_Data(iminus1_neighbour,1)
        yp_BP_iminus1 = BP_xyz_Data(iminus1_neighbour,2)
        zp_BP_iminus1 = BP_xyz_Data(iminus1_neighbour,3)
        xp_BP_iplus1 = BP_xyz_Data(iplus1_neighbour,1)
        yp_BP_iplus1 = BP_xyz_Data(iplus1_neighbour,2)
        zp_BP_iplus1 = BP_xyz_Data(iplus1_neighbour,3)

        xp_BP_jminus1 = BP_xyz_Data(jminus1_neighbour,1)
        yp_BP_jminus1 = BP_xyz_Data(jminus1_neighbour,2)
        zp_BP_jminus1 = BP_xyz_Data(jminus1_neighbour,3)
        xp_BP_jplus1 = BP_xyz_Data(jplus1_neighbour,1)
        yp_BP_jplus1 = BP_xyz_Data(jplus1_neighbour,2)
        zp_BP_jplus1 = BP_xyz_Data(jplus1_neighbour,3)
        
        if(i_absolute.eq.iminus1_neighbour)then
          xt_prev = xp_i - (xp_BP_iplus1 - xp_i)
          yt_prev = yp_i - (yp_BP_iplus1 - yp_i)
          zt_prev = zp_i - (zp_BP_iplus1 - zp_i)
          xt_next = xp_BP_iplus1
          yt_next = yp_BP_iplus1
          zt_next = zp_BP_iplus1
        else if(i_absolute.eq.iplus1_neighbour)then
          xt_prev = xp_BP_iminus1
          yt_prev = yp_BP_iminus1
          zt_prev = zp_BP_iminus1
          xt_next = xp_i + (xp_i - xp_BP_iminus1)
          yt_next = yp_i + (yp_i - yp_BP_iminus1)
          zt_next = zp_i + (zp_i - zp_BP_iminus1)
        else
          xt_prev = xp_BP_iminus1
          yt_prev = yp_BP_iminus1
          zt_prev = zp_BP_iminus1
          xt_next = xp_BP_iplus1
          yt_next = yp_BP_iplus1
          zt_next = zp_BP_iplus1
        end if
        if(i_absolute.eq.jminus1_neighbour)then
          xs_prev = xp_i - (xp_BP_jplus1 - xp_i)
          ys_prev = yp_i - (yp_BP_jplus1 - yp_i)
          zs_prev = zp_i - (zp_BP_jplus1 - zp_i)
          xs_next = xp_BP_jplus1
          ys_next = yp_BP_jplus1
          zs_next = zp_BP_jplus1
        else if(i_absolute.eq.jplus1_neighbour)then
          xs_prev = xp_BP_jminus1
          ys_prev = yp_BP_jminus1
          zs_prev = zp_BP_jminus1
          xs_next = xp_i + (xp_i - xp_BP_jminus1)
          ys_next = yp_i + (yp_i - yp_BP_jminus1)
          zs_next = zp_i + (zp_i - zp_BP_jminus1)
        else
          xs_prev = xp_BP_jminus1
          ys_prev = yp_BP_jminus1
          zs_prev = zp_BP_jminus1
          xs_next = xp_BP_jplus1
          ys_next = yp_BP_jplus1
          zs_next = zp_BP_jplus1
        end if
        
c        xtb,ytb,ztb are components of tangential
c       
        dxt_fwd = xt_next - xp_i
        dyt_fwd = yt_next - yp_i
        dzt_fwd = zt_next - zp_i
        dxt_back = (xp_i - xt_prev)
        dyt_back = (yp_i - yt_prev)
        dzt_back = (zp_i - zt_prev)

c        call periodicityCheck3D(dxt_fwd,  dyt_fwd , dyt_fwd ,
c     &                 xp_i, xt_next, yp_i, yt_next, zp_i, zt_next)
c        call periodicityCheck3D(dxt_back, dyt_back, dyt_back,
c     &                 xp_i, xt_prev, yp_i, yt_prev, zp_i, zt_prev)
        dxt = dxt_fwd + dxt_back
        dyt = dyt_fwd + dyt_back
        dzt = dzt_fwd + dzt_back

c        xsb,ysb,zsb are components of orthogonal tangential

c              
        dxs_fwd = xs_next - xp_i
        dys_fwd = ys_next - yp_i
        dzs_fwd = zs_next - zp_i
        dxs_back = (xp_i - xs_prev)
        dys_back = (yp_i - ys_prev)
        dzs_back = (zp_i - zs_prev)
        
c        call periodicityCheck3D(dxs_fwd,  dys_fwd , dys_fwd ,
c     &                 xp_i, xs_next, yp_i, ys_next, zp_i, zs_next)
c        call periodicityCheck3D(dxs_back, dys_back, dys_back,
c     &                 xp_i, xs_prev, yp_i, ys_prev, zp_i, zs_prev)
        dxs = dxs_fwd + dxs_back
        dys = dys_fwd + dys_back
        dzs = dzs_fwd + dzs_back

c       normal = vec(drt) x vec(drs)     :vector cross-product

        dxt_c = dxt_fwd + dxt_back
        dyt_c = dyt_fwd + dyt_back
        dzt_c = dzt_fwd + dzt_back
        dxs_c = dxs_fwd + dxs_back
        dys_c = dys_fwd + dys_back
        dzs_c = dzs_fwd + dzs_back

        dxn = dyt_c*dzs_c - dzt_c*dys_c   !dyt*dzr - dzt*dyr
        dyn = dzt_c*dxs_c - dxt_c*dzs_c   !dzt*dxr - dxt*dzr
        dzn = dxt_c*dys_c - dyt_c*dxs_c   !dxt*dyr - dyt*dxr
        
        drt = sqrt(dxt*dxt + dyt*dyt + dzt*dzt)
        drs = sqrt(dxs*dxs + dys*dys + dzs*dzs)
        drn = sqrt(dxn*dxn + dyn*dyn + dzn*dzn)

        if(.NOT.(drt.gt.0.0))then   !if(.NOT.(drn.gt.0.0))then
          print*
          print*,'ERROR in updateNormals_MPI_3D.f'
          !print*,'drn.eq.0 '
          print*,'drt.eq.0 '
          print*,'drn ',drn
          print*,'rank_id, itime ',rank_id, itime
          print*,'i ',i
          print*,'xp  , yp  , zp   ',xp(i),yp(i),zp(i)
          print*,'xp_i, yp_i, zp_i ',xp_i, yp_i, zp_i
          print*,'i_absolute ',i_absolute
          print*,'iminus1_neighbour ,iplus1_neighbour ',
     &            iminus1_neighbour ,iplus1_neighbour
          print*,'xt_next, yt_next, zt_next ',xt_next, yt_next, zt_next
          print*,'xt_prev, yt_prev, zt_prev ',xt_prev, yt_prev, zt_prev
          print*,'jminus1_neighbour, jplus1_neighbour ',
     &            jminus1_neighbour, jplus1_neighbour
          print*,'xs_next, ys_next, zs_next ',xs_next, ys_next, zs_next
          print*,'xs_prev, ys_prev, zs_prev ',xs_prev, ys_prev, zs_prev
          print*,'dxt_fwd , dyt_fwd , dzt_fwd  ',
     &            dxt_fwd , dyt_fwd , dzt_fwd
          print*,'dxt_back, dyt_back, dzt_back ',
     &            dxt_back, dyt_back, dzt_back
          print*,'dxt, dyt, dzt ',dxt, dyt, dzt
          print*,'dxs_fwd , dys_fwd , dzs_fwd  ',
     &            dxs_fwd , dys_fwd , dzs_fwd
          print*,'dxs_back, dys_back, dzs_back ',
     &            dxs_back, dys_back, dzs_back
          print*,'dxs, dys, dzs ',dxt, dys, dzs
          print*,'dxt_c, dyt_c, dzt_c ',dxt_c, dyt_c, dzt_c
          print*,'dxs_c, dys_c, dzs_c ',dxt_c, dys_c, dzs_c
          print*,'dxn, dyn, dzn ',dxn, dyn, dzn
          print*,'drn, drt, drs ',drn, drt, drs
          print*,'xtb, ytb, ztb ',xtb(i), ytb(i), ztb(i)
          print*,'xsb, ysb, zsb ',xsb(i), ysb(i), zsb(i)
          stop
        endif

        xtb(i) = dxt/drt
        ytb(i) = dyt/drt
        ztb(i) = dzt/drt
        
        xsb(i) = dxs/drs
        ysb(i) = dys/drs
        zsb(i) = dzs/drs

        
c       normal = vec(drt) x vec(drs)     :vector cross-product
        xnb(i) = dxn/drn
        ynb(i) = dyn/drn
        znb(i) = dzn/drn

        deltaptb(i,1) = sqrt((dxt_back)**2 + (dyt_back)**2 
     &                     + (dzt_back)**2)
        deltaptb(i,2) = sqrt((dxt_fwd)**2 + (dyt_fwd)**2
     &                     + (dzt_fwd)**2)
        deltapsb(i,1) = sqrt((dxs_back)**2 + (dys_back)**2 
     &                     + (dzs_back)**2)
        deltapsb(i,2) = sqrt((dxs_fwd)**2 + (dys_fwd)**2
     &                     + (dzs_fwd)**2)

      
      enddo


      end subroutine
