c    "Copyright 2010 Prof. Robert Dalrymple, Prof. M. Gomez Gesteira, Dr Benedict Rogers, 
c     Dr Alejandro Crespo, Dr Muthukumar Narayanaswamy, Dr Shan Zou, Dr Andrea Panizzo "
c
c    This file is part of SPHYSICS.
c
c    SPHYSICS is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 3 of the License, or
c    (at your option) any later version.
c
c    SPHYSICS is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program.  If not, see <http://www.gnu.org/licenses/>.

      subroutine variable_time_step

      include 'common.3D'
      
      real dt_new_temp(n_procs_max)

C     Calculate time step 
c     Conditions given in 
c     Monaghan (1989), JCP 82, 1-15
c     Monaghan & Kos (1999), JWPCOE 125, 145-154

      Fa_max=-1000.0
      cs_max=-1000.0
      i_fa = -1000

      dt_new_temp = 1000.0
       
      if(iBC.eq.1)then
        nstart_VTS = nstart
      else
        nstart_VTS = nbp1_local
      endif

       do ipcount=nstart_VTS,np_local
          call ipcount_to_i(ipcount,i)
	  Fa=sqrt(ax(i)*ax(i)+ay(i)*ay(i)+az(i)*az(i))
	  
	  
	  if(Fa.gt.Fa_max)then
           i_fa=i
           Fa_max=Fa
	  endif
          if(cs(i).gt.cs_max)then
           cs_max=cs(i)
          endif
	enddo 

      if(num_FB.gt.0)then
       do ipcount = nbfm_local+1,nbfreem_local   !do i=nbfm+1,nb
        i = ip_index_BP_local_3(ipcount-nbfm_local)
        i_absolute = iabsolute_BP_Pointer_Index(i)
        i_num_FB = i_FB_Pointer_Info(i_absolute)
        Fa=sqrt((ax(i) )**2 + ay(i)**2
     &        + (az(i) )**2) / pm(i)

        if (Fa.ge.Fa_max) then
           i_fa=i
           Fa_max=Fa
        endif
        if(cs(i).ge.cs_max)then
          cs_max=cs(i)
        endif
        vel_magnitude = sqrt(up(i)*up(i) + vp(i)*vp(i) + wp(i)*wp(i))
        if (vel_magnitude.ge.cs_max) then
           cs_max=vel_magnitude
        endif
       enddo
       if(.NOT.(abs(Fa_max).gt.0.0))then
         i_fa = 0
       endif
      !print*,'i_fa, Fa_max ',i_fa, Fa_max
      endif    !End of:   if(num_FB.gt.0)then

        if(i_fa.gt.0)then
          Fa_sqrt=((ax(i_fa)+grx)**2 + (ay(i_fa)+gry)**2
     &            +(az(i_fa)+grz)**2)**0.25
          
          dt_1=sqrt(h)/Fa_sqrt
        else
          Fa_sqrt=((grx)**2 + (gry)**2 + (grz)**2)**0.25
          dt_1=sqrt(h)/Fa_sqrt
        endif

       !dt_2=h/(cs_max+h*visc_dt)
       if(cs_max.gt.0.0)then      
         dt_2=h/(cs_max+h*visc_dt)       
       else       
         dt_2=dt_1         
       endif
       
       dt_new_irank=CFL_number*min(dt_1,dt_2)
       
!       !- Debugging info -
!       if(rank_id.eq.4.or.rank_id.eq.10)then       
!        print*,'rank_id, itime, dt_1,dt_2,dt_new,i_fa,iPT,Fa_sqrt,'
!      &        //'ax, ay, az, cs_max '
!        print*,rank_id,itime, dt_1,dt_2,dt_new_irank,i_fa,
!      &    i_particleType(i),Fa_sqrt,ax(i_fa),ay(i_fa),az(i_fa),cs_max,
!      &    xp(i_fa),yp(i_fa),zp(i_fa)
! !        if(i_fa.gt.0)then   
! !          write(*,455)rank_id,itime, dt_1,dt_2,dt_new_irank,i_fa,Fa_sqrt,
! !      &          ax(i_fa),az(i_fa),cs_max 
! !        else
! !          write(*,455)rank_id,itime, dt_1,dt_2,dt_new_irank,i_fa,Fa_sqrt,
! !      &          cs_max 
! !        endif 
! ! 455    format(i3,i6,3(1x,f12.6),i6,4(1x,e12.6))       
!        print*,'CFL_number, rank_id ',CFL_number, rank_id
!        print*,'nstart_VTS, np_local ',nstart_VTS,np_local
!       endif
       
       !-- gather and compare min dt from each processor --
       call MPI_GATHER(dt_new_irank          , 1, MPI_REAL, 
     &                 dt_new_temp(rank_id+1), 1, MPI_REAL, 
     &                 0, MPI_COMM_WORLD, ierr)
     
       if(rank_id.eq.0)dt_new = minval(dt_new_temp)
      
       call MPI_BCast(dt_new, 1, MPI_REAL, 
     &                  0, MPI_COMM_WORLD, ierr)
       
       
	return

	end

!       if(rank_id.eq.0)then      
!        print* 
!        print*,'rank_id, itime ',rank_id, itime
!        do i_procs=1,n_procs
!          print*,'i_procs, dt_new_temp ',i_procs-1, dt_new_temp(i_procs)
!        enddo
!        print*,'dt_new = ',dt_new
!        if(itime.eq.1)stop
!       endif

